#####################################################################################
#
#  Copyright (c) Crossbar.io Technologies GmbH
#  SPDX-License-Identifier: EUPL-1.2
#
#####################################################################################

import sys
import importlib

from autobahn.twisted.wamp import ApplicationSession
from autobahn.wamp.exception import ApplicationError

from twisted.python.failure import Failure

from txaio import make_logger

#
# the imports here are important (though not used in CB unless configured),
# because of single-exe packaging and pyinstaller otherwise missing deps
#

from crossbar.worker.sample import LogTester  # noqa


def _appsession_loader(config):
    """
    Load a class from C{config}.
    """
    log = make_logger()

    if config['type'] == 'class':

        try:
            klassname = config['classname']

            log.debug("Starting class '{klass}'", klass=klassname)

            c = klassname.split('.')
            module_name, klass_name = '.'.join(c[:-1]), c[-1]
            module = importlib.import_module(module_name)
            component = getattr(module, klass_name)

            if not issubclass(component, ApplicationSession):
                raise ApplicationError("crossbar.error.class_import_failed",
                                       "session not derived of ApplicationSession")

        except Exception:
            emsg = "Failed to import class '{}'\n{}".format(klassname, Failure().getTraceback())
            log.debug(emsg)
            log.debug("PYTHONPATH: {pythonpath}", pythonpath=sys.path)
            raise ApplicationError("crossbar.error.class_import_failed", emsg, pythonpath=sys.path)

    elif config['type'] == 'function':
        callbacks = {}
        for name, funcref in config.get('callbacks', {}).items():
            if '.' not in funcref:
                raise ApplicationError(
                    "crossbar.error",
                    "no '.' in callback reference '{}'".format(funcref),
                )

            try:
                package, func = funcref.rsplit('.', 1)

                module = importlib.import_module(package)
                callbacks[name] = getattr(module, func)

            except Exception:
                emsg = "Failed to import package '{}' (for '{}')\n{}".format(package, funcref,
                                                                             Failure().getTraceback())
                log.error('{msg}', msg=emsg)
                raise ApplicationError("crossbar.error.class_import_failed", emsg)

        # while the "component" callback is usually an
        # ApplicationSession class, it can be anything that takes a
        # "config" arg (must return an ApplicationSession instance)
        def component(cfg):
            session = _AnonymousRoleSession(cfg)
            session.role = config.get('role', 'anonymous')
            for name, fn in callbacks.items():
                session.on(name, fn)
            return session

    else:
        raise ApplicationError("crossbar.error.invalid_configuration",
                               "invalid component type '{}'".format(config['type']))

    return component


class _AnonymousRoleSession(ApplicationSession):

    role = None

    def onConnect(self):
        self.join(self.config.realm, authrole=self.role)



##############################################################################
#
#                        Crossbar.io
#     Copyright (C) Crossbar.io Technologies GmbH. All rights reserved.
#
##############################################################################

import pprint

import numpy as np

from twisted.internet.defer import inlineCallbacks, succeed
from twisted.internet.task import LoopingCall

from autobahn.wamp.types import SubscribeOptions, PublishOptions
from autobahn.wamp.exception import ApplicationError, TransportLost
from autobahn.twisted.wamp import ApplicationSession

from crossbar._util import hlid, hl, hltype

import txaio
from txaio import make_logger, time_ns

__all__ = ('NodeManagementSession', 'NodeManagementBridgeSession')


class NodeManagementSession(ApplicationSession):
    """
    This session is used for the uplink connection to
    Crossbar.io Master.
    """

    log = make_logger()

    def onUserError(self, fail, msg):
        """
        Implements :func:`autobahn.wamp.interfaces.ISession.onUserError`
        """
        if isinstance(fail.value, ApplicationError):
            self.log.debug('{klass}.onUserError(): "{msg}"',
                           klass=self.__class__.__name__,
                           msg=fail.value.error_message())
        else:
            self.log.error(
                '{klass}.onUserError(): "{msg}"\n{traceback}',
                klass=self.__class__.__name__,
                msg=msg,
                traceback=txaio.failure_format_traceback(fail),
            )

    def __init__(self, runner, config=None):
        ApplicationSession.__init__(self, config)
        self._runner = runner

    def onConnect(self):
        self.log.debug('{klass}.onConnect()', klass=self.__class__.__name__)

        # we're connected, now authenticate using wamp-cryptosign ..

        extra = {
            # forward the client pubkey: this allows us to omit authid as
            # the router can identify us with the pubkey already
            u'pubkey': self.config.extra['node_key'].public_key(),

            # not yet implemented. a public key the router should provide
            # a trustchain for it's public key. the trustroot can eg be
            # hard-coded in the client, or come from a command line option.
            u'trustroot': None,

            # not yet implemented. for authenticating the router, this
            # challenge will need to be signed by the router and send back
            # in AUTHENTICATE for client to verify. A string with a hex
            # encoded 32 bytes random value.
            u'challenge': None,

            # https://tools.ietf.org/html/rfc5929
            u'channel_binding': u'tls-unique'
        }

        # now request to join. the authrole==node is mandatory. the actual realm
        # we're joined to is decided by Crossbar.io Master, and hence we
        # must not provide that. Same holds for authid (also auto-assigned).
        # the authid assigned will (also) be used as the node_id
        self.join(realm=None, authrole=u'node', authmethods=[u'cryptosign'], authextra=extra)

    def onChallenge(self, challenge):
        self.log.debug('{klass}.onChallenge(challenge={challenge})',
                       klass=self.__class__.__name__,
                       challenge=challenge)

        if challenge.method == u'cryptosign':
            # alright, we've got a challenge from the router.

            # not yet implemented. check the trustchain the router provided against
            # our trustroot, and check the signature provided by the
            # router for our previous challenge. if both are ok, everything
            # is fine - the router is authentic wrt our trustroot.

            # sign the challenge with our private key.
            signed_challenge = self.config.extra['node_key'].sign_challenge(self, challenge)

            # send back the signed challenge for verification
            return signed_challenge

        else:
            raise Exception(
                'internal error: we asked to authenticate using wamp-cryptosign, but now received a challenge for {}'.
                format(challenge.method))

    def onJoin(self, details):
        self.log.info('{func}(details={details})', func=hltype(self.onJoin), details=details)

        # be paranoid .. sanity checks
        if self.config.extra and u'on_ready' in self.config.extra:
            if not self.config.extra[u'on_ready'].called:

                self.config.extra[u'on_ready'].callback((
                    self,
                    details.realm,  # the management realm we've got auto-assigned to
                    details.session,  # WAMP session ID
                    details.authid,  # the authid (==node_id) we've got auto-assigned
                    details.authextra))
            else:
                raise Exception('internal error: on_ready callback already called when we expected it was not')
        else:
            raise Exception('internal error: no on_ready callback provided')

    def onLeave(self, details):
        self.log.debug('{klass}.onLeave(details={details})', klass=self.__class__.__name__, details=details)

        if details.reason in ['fabric.auth-failed.node-unpaired', 'fabric.auth-failed.node-already-connected']:
            # no reason to auto-reconnect: user needs to get active and pair the node first.
            self._runner.stop()

        if self.config.extra and u'on_ready' in self.config.extra:
            if not self.config.extra[u'on_ready'].called:
                self.config.extra['on_ready'].errback(ApplicationError(details.reason, details.message))

        if self.config.extra and u'on_exit' in self.config.extra:
            if not self.config.extra[u'on_exit'].called:
                self.config.extra['on_exit'].callback(details.reason)
            else:
                raise Exception('internal error: on_exit callback already called when we expected it was not')
        else:
            raise Exception('internal error: no on_exit callback provided')

        self.disconnect()

    def onDisconnect(self):
        self.log.debug('{klass}.onDisconnect()', klass=self.__class__.__name__)

        node = self.config.extra[u'node']

        # FIXME: the node shutdown behavior should be more sophisticated than this!
        shutdown_on_cfc_lost = False

        if shutdown_on_cfc_lost:
            if node._controller:
                node._controller.shutdown()


class NodeManagementBridgeSession(ApplicationSession):
    """
    The management bridge is a WAMP session that lives on the local management router,
    but has access to a 2nd WAMP session that lives on the uplink CFC router.

    The bridge is responsible for forwarding calls from CFC into the local node,
    and for forwarding events from the local node to CFC.
    """

    log = make_logger()

    def __init__(self, config):
        ApplicationSession.__init__(self, config)
        self._manager = None
        self._management_realm = None
        self._node_id = None
        self._regs = {}
        self._authrole = u'trusted'
        self._authmethod = u'trusted'
        self._sub_on_mgmt = None
        self._sub_on_reg_create = None
        self._sub_on_reg_delete = None
        self._heartbeat_time_ns = None
        self._heartbeat = 0
        self._heartbeat_call = None

    def onJoin(self, details):
        self.log.debug('{klass}.onJoin(details={details})', klass=self.__class__.__name__, details=details)

    @inlineCallbacks
    def attach_manager(self, manager, management_realm, node_id):
        """
        Attach management uplink session when the latter has been fully established
        and is ready to be used.

        :param manager: uplink session.
        :type manager: instance of `autobahn.wamp.protocol.ApplicationSession`

        :param management_realm: The management realm that was assigned by CFC to this node.
        :type management_realm: unicode

        :param node_id: The node ID that was assigned by CFC to this node.
        :type node_id: unicode
        """
        if self._manager:
            raise Exception('{}.attach_manager: interal error: manager already attached!'.format(
                self.__class__.__name__))

        self._manager = manager
        self._management_realm = management_realm
        self._node_id = node_id
        self._node_key = self.config.extra['node_key']
        self._controller_config = self.config.extra['controller_config']

        fabric = self._controller_config.get('fabric', {})
        heartbeat = fabric.get('heartbeat', {})
        self._heartbeat_startup_delay = heartbeat.get('startup_delay', 5)
        self._heartbeat_heartbeat_period = heartbeat.get('heartbeat_period', 10)
        self._heartbeat_include_system_stats = heartbeat.get('include_system_stats', True)
        self._heartbeat_send_workers_heartbeats = heartbeat.get('send_workers_heartbeats', True)
        self._heartbeat_aggregate_workers_heartbeats = heartbeat.get('aggregate_workers_heartbeats', True)

        yield self._start_call_forwarding()
        yield self._start_event_forwarding()

        # start heartbeating a bit later (we currently run into workers not fully being up otherwise)
        from twisted.internet import reactor
        reactor.callLater(self._heartbeat_startup_delay, self._start_cfc_heartbeat)

        self.log.info(
            '{klass}.attach_manager: manager attached as node "{node_id}" on management realm "{management_realm}") with public key "{public_key}"',
            klass=self.__class__.__name__,
            node_id=self._node_id,
            management_realm=self._management_realm,
            public_key=self._node_key.public_key())
        self.log.info('Controller configuration: {controller_config}', controller_config=self._controller_config)

    @inlineCallbacks
    def detach_manager(self):
        """
        Detach management uplink session (eg when that session has been lost).
        """
        if not self._manager:
            self.log.debug('{klass}.detach_manager: no manager manager currently attached',
                           klass=self.__class__.__name__)
            return

        try:
            self._stop_cfc_heartbeat()
        except:  # noqa
            self.log.failure()

        try:
            yield self._stop_event_forwarding()
        except:  # noqa
            self.log.failure()

        try:
            yield self._stop_call_forwarding()
        except:  # noqa
            self.log.failure()

        self._manager = None
        self._management_realm = None
        self._node_id = None

        self.log.info(
            '{klass}.detach_manager: manager detached for node "{node_id}" on management realm "{management_realm}")',
            klass=self.__class__.__name__,
            node_id=self._node_id,
            management_realm=self._management_realm)

    def _translate_uri(self, uri):
        """
        Translate a local URI (one that is used on the local node management router)
        to a remote URI (one used on the uplink management session at the CFC router
        for the management realm).

        Example:

            crossbar.worker.worker-001.start_manhole
                ->
            crossbarfabriccenter.node.<node_id>.worker.<worker_id>.start_manhole
        """

        # the local URI prefix under which the management API is registered
        _PREFIX = 'crossbar.'

        # the remote (==CFC) URI prefix under which the management API is registered
        _TARGET_PREFIX = u'crossbarfabriccenter.node'

        if uri.startswith(_PREFIX):
            suffix = uri[len(_PREFIX):]
            mapped_uri = u'.'.join([_TARGET_PREFIX, self._node_id, suffix])
            self.log.debug("mapped URI {uri} to {mapped_uri} [suffix={suffix}]",
                           uri=uri,
                           mapped_uri=mapped_uri,
                           suffix=suffix)
            return mapped_uri

        raise Exception("don't know how to translate URI {}".format(uri))

    @inlineCallbacks
    def _send_heartbeat(self):
        def _drop_attr(status):
            for k in ['ts', 'timestamp', 'user', 'name', 'cmdline', 'created']:
                if k in status:
                    del status[k]

        if self._manager and self._manager.is_attached():
            node_pubkey = str(self._node_key.public_key())

            # get basic status
            status = yield self.call('crossbar.get_status')
            obj = {
                'timestamp': self._heartbeat_time_ns,
                'period': self._heartbeat_heartbeat_period,
                'pubkey': node_pubkey,
                'mrealm_id': self._management_realm,
                'seq': self._heartbeat,
                'workers': status.get('workers_by_type', {}),
            }

            controller_status = None
            try:
                controller_status = yield self.call('crossbar.get_process_monitor')
            except ApplicationError as e:
                if e.error == 'wamp.error.no_such_procedure':
                    self.log.info(
                        'Failed to retrieve controller process statistics for period {period} - controller procedure unavailable',
                        period=self._heartbeat)
                else:
                    raise e
            else:
                _drop_attr(controller_status)
                controller_status['timestamp'] = self._heartbeat_time_ns
                controller_status['period'] = self._heartbeat_heartbeat_period
                controller_status['mrealm_id'] = self._management_realm
                controller_status['seq'] = self._heartbeat
                controller_status['type'] = 'controller'

                # MWorkerState: "online and fully operational" == 1
                controller_status['state'] = 1

            obj['workers']['controller'] = 1

            if self._heartbeat_include_system_stats:
                obj['system'] = yield self.call('crossbar.get_system_stats')

            # MNodeState: "online and fully operational" == 1
            obj['state'] = 1

            try:
                if self._manager:
                    yield self._manager.publish(u'crossbarfabriccenter.node.on_heartbeat',
                                                self._node_id,
                                                obj,
                                                options=PublishOptions(acknowledge=True))
                    self.log.debug('Node heartbeat sent [node_id="{node_id}", timestamp="{timestamp}", seq={seq}]',
                                   timestamp=np.datetime64(obj['timestamp'], 'ns'),
                                   seq=obj['seq'],
                                   node_id=self._node_id,
                                   obj=obj)
                    self.log.debug('{heartbeat}', heartbeat=pprint.pformat(obj))
                else:
                    self.log.warn(
                        'Skipped sending management link node heartbeat for period {period} - no management uplink',
                        period=self._heartbeat)
            except:  # noqa
                self.log.warn('Failed to send management link node heartbeat for period {period}:',
                              period=self._heartbeat)
                self.log.failure()

            if self._heartbeat_send_workers_heartbeats:
                workers = {}
                native_worker_types = ['router', 'container', 'proxy', 'xbrmm', 'hostmonitor']
                worker_ids = yield self.call('crossbar.get_workers', filter_types=native_worker_types)
                for worker_id in worker_ids:
                    worker_status = {}
                    try:
                        worker_status['process'] = yield self.call(
                            'crossbar.worker.{}.get_process_monitor'.format(worker_id))
                        self.log.debug('Native worker status: {status}', status=worker_status)
                    except ApplicationError as e:
                        if e.error == 'wamp.error.no_such_procedure':
                            self.log.warn(
                                'Failed to retrieve worker process statistics for worker "{worker_id}" in period {period} ("worker procedure unavailable")',
                                worker_id=worker_id,
                                period=self._heartbeat)
                        else:
                            raise e
                    else:
                        _drop_attr(worker_status)
                        worker_status['timestamp'] = self._heartbeat_time_ns
                        worker_status['period'] = self._heartbeat_heartbeat_period
                        worker_status['pubkey'] = node_pubkey
                        worker_status['mrealm_id'] = self._management_realm
                        worker_status['seq'] = self._heartbeat
                        worker_status['type'] = worker_status['process']['type']

                        # MWorkerState: "online and fully operational" == 1
                        worker_status['state'] = 1

                        del worker_status['process']['type']
                        workers[worker_id] = worker_status

                workers['controller'] = controller_status

                for worker_id, worker_status in workers.items():

                    # if worker is of type "router", expand with router statistics
                    if worker_status['type'] == 'router':
                        try:
                            router_stats = yield self.call(
                                'crossbar.worker.{}.get_router_realm_stats'.format(worker_id))
                            self.log.debug('Router worker status: {router_stats}', router_stats=router_stats)
                        except ApplicationError as e:
                            if e.error == 'wamp.error.no_such_procedure':
                                self.log.warn(
                                    'Failed to retrieve router statistics for period {period} - worker procedure unavailable',
                                    period=self._heartbeat)
                            else:
                                raise e
                        else:
                            if self._heartbeat_aggregate_workers_heartbeats:
                                mstats = {}
                                stats_sessions = 0
                                stats_roles = 0
                                for realm in router_stats:
                                    stats_sessions += router_stats[realm]['sessions']
                                    stats_roles += router_stats[realm]['roles']
                                    for direction in router_stats[realm]['messages']:
                                        for k in router_stats[realm]['messages'][direction]:
                                            if k not in mstats:
                                                mstats[k] = 0
                                            mstats[k] += router_stats[realm]['messages'][direction][k]
                                worker_status['router'] = {
                                    'messages': mstats,
                                    'sessions': stats_sessions,
                                    'roles': stats_roles,
                                }
                            else:
                                worker_status['router'] = router_stats

                    # now publish the node heartbeat management event
                    try:
                        if self._manager:
                            yield self._manager.publish(u'crossbarfabriccenter.node.on_worker_heartbeat',
                                                        self._node_id,
                                                        str(worker_id),
                                                        worker_status,
                                                        options=PublishOptions(acknowledge=True))
                            self.log.debug(
                                'Worker heartbeat sent [node_id="{node_id}", worker_id="{worker_id}", timestamp="{timestamp}", seq={seq}]',
                                node_id=self._node_id,
                                worker_id=worker_id,
                                timestamp=np.datetime64(worker_status['timestamp'], 'ns'),
                                seq=worker_status['seq'])
                            self.log.debug('{heartbeat}', heartbeat=pprint.pformat(worker_status))
                        else:
                            self.log.warn(
                                'Skipped sending management link node heartbeat for period {period} - no management uplink',
                                period=self._heartbeat)
                    except TransportLost:
                        self.log.info(
                            'Failed to send management link worker heartbeat for period {period} - transport lost',
                            period=self._heartbeat)

        else:
            self.log.info('Skipped sending management link heartbeat for period {period} (not connected)',
                          period=self._heartbeat)
            return succeed(None)

    def _start_cfc_heartbeat(self):
        self.log.info('Starting management heartbeat .. [period={period} seconds]',
                      period=hlid(self._heartbeat_heartbeat_period))

        self._heartbeat_time_ns = None
        self._heartbeat = 0

        @inlineCallbacks
        def publish():
            self._heartbeat_time_ns = time_ns()
            self._heartbeat += 1
            try:
                yield self._send_heartbeat()
            except Exception:
                self.log.failure()

        self._heartbeat_call = LoopingCall(publish)
        self._heartbeat_call.start(self._heartbeat_heartbeat_period)

    def _stop_cfc_heartbeat(self):
        if self._heartbeat_call:
            self._heartbeat_call.stop()
            self._heartbeat_call = None
            self._heartbeat_time_ns = None
            self._heartbeat = 0

    @inlineCallbacks
    def _start_event_forwarding(self):

        # setup event forwarding (events originating locally are forwarded uplink)
        #
        @inlineCallbacks
        def on_management_event(*args, **kwargs):
            if not (self._manager and self._manager.is_attached()):
                self.log.warn("Can't foward management event: CFC session not attached")
                return

            details = kwargs.pop('details')

            # a node local event such as 'crossbar.node.on_ready' is mogrified to 'local.crossbar.node.on_ready'
            # (one reason is that URIs such as 'wamp.*' and 'crossbar.*' are restricted to trusted sessions, and
            # the management bridge is connecting over network to the uplink CFC and hence can't be trusted)
            #
            topic = self._translate_uri(details.topic)

            try:
                yield self._manager.publish(topic, *args, options=PublishOptions(acknowledge=True), **kwargs)
            except Exception:
                self.log.failure(
                    "Failed to forward event on topic '{topic}': {log_failure.value}",
                    topic=topic,
                )
            else:
                if topic.endswith('.on_log'):
                    log = self.log.debug
                else:
                    log = self.log.debug

                log('Forwarded management {forward_type} to CFC [local_uri={local_topic}, remote_uri={remote_topic}]',
                    forward_type=hl('EVENT'),
                    local_topic=hlid(details.topic),
                    remote_topic=hlid(topic))

        try:
            sub = self._sub_on_mgmt = yield self.subscribe(on_management_event,
                                                           u"crossbar.",
                                                           options=SubscribeOptions(match=u"prefix",
                                                                                    details_arg="details"))
            self.log.debug("Setup prefix subscription to forward node management events: {sub}", sub=sub)
        except:
            self.log.failure()

    @inlineCallbacks
    def _stop_event_forwarding(self):
        if self._sub_on_mgmt:
            yield self._sub_on_mgmt.unsubscribe()
            self._sub_on_mgmt = None

    @inlineCallbacks
    def _start_call_forwarding(self):

        # forward future new registrations
        #
        @inlineCallbacks
        def on_registration_create(session_id, registration):
            # we use the WAMP meta API implemented by CB to get notified whenever a procedure is
            # registered/unregister on the node management router, setup a forwarding procedure
            # and register that on the uplink CFC router

            if not (self._manager and self._manager.is_attached()):
                self.log.warn("Can't create forward management registration: CFC session not attached")
                return

            local_uri = registration['uri']
            remote_uri = self._translate_uri(local_uri)

            self.log.debug('Setup management API forwarding: {remote_uri} -> {local_uri}',
                           remote_uri=remote_uri,
                           local_uri=local_uri)

            def forward_call(*args, **kwargs):
                kwargs.pop('details', None)
                self.log.debug(
                    'Forwarding management {forward_type} from CFC .. [remote_uri={remote_uri}, local_uri={local_uri}]',
                    forward_type=hl('CALL'),
                    local_uri=hlid(local_uri),
                    remote_uri=hlid(remote_uri))
                return self.call(local_uri, *args, **kwargs)

            try:
                reg = yield self._manager.register(forward_call, remote_uri)
            except Exception:
                self.log.failure(
                    "Failed to register management procedure '{remote_uri}': {log_failure.value}",
                    remote_uri=remote_uri,
                )
            else:
                self._regs[registration['id']] = reg
                self.log.debug("Management procedure registered: '{remote_uri}'", remote_uri=reg.procedure)

        self._sub_on_reg_create = yield self.subscribe(on_registration_create, u'wamp.registration.on_create')

        # stop forwarding future registrations
        #
        @inlineCallbacks
        def on_registration_delete(session_id, registration_id):
            if not (self._manager and self._manager.is_attached()):
                self.log.debug("Can't delete forward management registration: CFC session not attached")
                return

            reg = self._regs.pop(registration_id, None)

            if reg:
                yield reg.unregister()
                self.log.debug("Management procedure unregistered: '{remote_uri}'", remote_uri=reg.procedure)
            else:
                self.log.warn("Could not remove forwarding for unmapped registration_id {reg_id}",
                              reg_id=registration_id)

        self._sub_on_reg_delete = yield self.subscribe(on_registration_delete, u'wamp.registration.on_delete')

        # start forwarding current registrations
        #
        res = yield self.call(u'wamp.registration.list')
        for match_type, reg_ids in res.items():
            for reg_id in reg_ids:
                registration = yield self.call(u'wamp.registration.get', reg_id)
                if registration[u'uri'].startswith(u'crossbar.'):
                    yield on_registration_create(None, registration)
                else:
                    # eg skip WAMP meta API procs like "wamp.session.list"
                    self.log.debug('skipped {uri}', uri=registration[u'uri'])

    @inlineCallbacks
    def _stop_call_forwarding(self):
        if self._sub_on_reg_create:
            yield self._sub_on_reg_create.unsubscribe()
            self._sub_on_reg_create = None

        if self._sub_on_reg_delete:
            yield self._sub_on_reg_delete.unsubscribe()
            self._sub_on_reg_delete = None



#####################################################################################
#
#  Copyright (c) Crossbar.io Technologies GmbH
#  SPDX-License-Identifier: EUPL-1.2
#
#####################################################################################

from os import getcwd, chdir

from twisted.trial.unittest import TestCase as _TestCase


class TestCase(_TestCase):
    """
    A Trial TestCase that makes sure that it is in the same directory when it
    finishes as when it began. This is because the CLI changes directories,
    meaning we can end up making thousands deep file structures.
    """
    def setUp(self):
        cb_original_dir = getcwd()
        self.addCleanup(lambda: chdir(cb_original_dir))
        return super(TestCase, self).setUp()



#####################################################################################
#
#  Copyright (c) Crossbar.io Technologies GmbH
#  SPDX-License-Identifier: EUPL-1.2
#
#####################################################################################

from twisted.internet.endpoints import TCP4ClientEndpoint
from twisted.internet.selectreactor import SelectReactor

from .interop import Result, ReplayClientFactory, Frame, ConnectionLoss
from crossbar.bridge.mqtt_events import (Connect, ConnectFlags, ConnACK, Publish, PubACK, PubREL, PubREC, PubCOMP,
                                         Subscribe, SubscriptionTopicRequest, SubACK, Disconnect)


def test_connect(host, port):
    record = [
        Frame(send=True, data=Connect(client_id="test_cleanconnect", flags=ConnectFlags(clean_session=True))),
        Frame(send=False, data=ConnACK(session_present=False, return_code=0)),
        Frame(send=True, data=Disconnect()),
        ConnectionLoss(),
    ]

    r = SelectReactor()
    f = ReplayClientFactory(r, record)
    e = TCP4ClientEndpoint(r, host, port)
    e.connect(f)
    r.run()

    return Result("connect", f.success, f.reason, f.client_transcript)


def test_quirks_mode_connect(host, port):
    record = [
        Frame(send=True, data=b"\x10\x15\x00\x04MQTT\x04\x02\x00x\x00\x07testqrk\x00\x00"),
        Frame(send=False, data=ConnACK(session_present=False, return_code=0)),
        Frame(send=True, data=Disconnect()),
        ConnectionLoss(),
    ]

    r = SelectReactor()
    f = ReplayClientFactory(r, record)
    e = TCP4ClientEndpoint(r, host, port)
    e.connect(f)
    r.run()

    return Result("connect_quirks", f.success, f.reason, f.client_transcript)


def test_reserved_packet_15(host, port):
    record = [
        Frame(send=True, data=Connect(client_id="test_reserved15", flags=ConnectFlags(clean_session=True))),
        Frame(send=False, data=ConnACK(session_present=False, return_code=0)),
        Frame(
            send=True,
            #        v pkt 15 right here
            data=b"\xf0\x13\x00\x04MQTT\x04\x02\x00\x02\x00\x07test123"),
        ConnectionLoss()
    ]

    r = SelectReactor()
    f = ReplayClientFactory(r, record)
    e = TCP4ClientEndpoint(r, host, port)
    e.connect(f)
    r.run()

    return Result("reserved_pkt15", f.success, f.reason, f.client_transcript)


def test_reserved_packet_0(host, port):
    record = [
        Frame(send=True, data=Connect(client_id="test_reserved0", flags=ConnectFlags(clean_session=True))),
        Frame(send=False, data=ConnACK(session_present=False, return_code=0)),
        Frame(
            send=True,
            #        v pkt 0 right here
            data=b"\x00\x13\x00\x04MQTT\x04\x02\x00\x02\x00\x07test123"),
        ConnectionLoss()
    ]

    r = SelectReactor()
    f = ReplayClientFactory(r, record)
    e = TCP4ClientEndpoint(r, host, port)
    e.connect(f)
    r.run()

    return Result("reserved_pkt0", f.success, f.reason, f.client_transcript)


def test_uninvited_puback(host, port):
    record = [
        Frame(send=True, data=Connect(client_id="test_puback", flags=ConnectFlags(clean_session=True))),
        Frame(send=False, data=ConnACK(session_present=False, return_code=0)),
        Frame(send=True, data=PubACK(packet_identifier=1234)),
        Frame(send=False, data=b""),
        Frame(send=True, data=Disconnect()),
        ConnectionLoss(),
    ]

    r = SelectReactor()
    f = ReplayClientFactory(r, record)
    e = TCP4ClientEndpoint(r, host, port)
    e.connect(f)
    r.run()

    return Result("uninvited_puback", f.success, f.reason, f.client_transcript)


def test_uninvited_pubrel(host, port):
    record = [
        Frame(send=True, data=Connect(client_id="test_pubrel", flags=ConnectFlags(clean_session=True))),
        Frame(send=False, data=ConnACK(session_present=False, return_code=0)),
        Frame(send=True, data=PubREL(packet_identifier=1234)),
        Frame(send=False, data=PubCOMP(packet_identifier=1234)),
        Frame(send=True, data=Disconnect()),
        ConnectionLoss(),
    ]

    r = SelectReactor()
    f = ReplayClientFactory(r, record)
    e = TCP4ClientEndpoint(r, host, port)
    e.connect(f)
    r.run()

    return Result("uninvited_pubrel", f.success, f.reason, f.client_transcript)


def test_self_subscribe(host, port):
    record = [
        Frame(send=True, data=Connect(client_id="test_selfsub", flags=ConnectFlags(clean_session=True))),
        Frame(send=False, data=ConnACK(session_present=False, return_code=0)),
        Frame(send=True, data=Subscribe(packet_identifier=1234, topic_requests=[SubscriptionTopicRequest("foo", 2)])),
        Frame(send=False, data=SubACK(packet_identifier=1234, return_codes=[2])),
        Frame(send=True, data=Publish(duplicate=False, qos_level=0, topic_name="foo", payload=b"abc", retain=False)),
        Frame(send=False, data=Publish(duplicate=False, qos_level=0, topic_name="foo", payload=b"abc", retain=False)),
        Frame(send=True, data=Disconnect()),
        ConnectionLoss(),
    ]

    r = SelectReactor()
    f = ReplayClientFactory(r, record)
    e = TCP4ClientEndpoint(r, host, port)
    e.connect(f)
    r.run()

    return Result("self_subscribe", f.success, f.reason, f.client_transcript)


def test_qos2_send_wrong_confirm(host, port):
    record = [
        Frame(send=True, data=Connect(client_id="test_wrong_confirm_qos2", flags=ConnectFlags(clean_session=True))),
        Frame(send=False, data=ConnACK(session_present=False, return_code=0)),
        Frame(send=True, data=Subscribe(packet_identifier=1234, topic_requests=[SubscriptionTopicRequest("foo", 2)])),
        Frame(send=False, data=SubACK(packet_identifier=1234, return_codes=[2])),
        Frame(send=True,
              data=Publish(duplicate=False,
                           qos_level=2,
                           topic_name="foo",
                           payload=b"abc",
                           retain=False,
                           packet_identifier=12)),
        Frame(send=False,
              data=[
                  PubREC(packet_identifier=12),
                  Publish(duplicate=False,
                          qos_level=2,
                          topic_name="foo",
                          payload=b"abc",
                          retain=False,
                          packet_identifier=1),
                  PubCOMP(packet_identifier=12)
              ]),
        Frame(send=True, data=PubREL(packet_identifier=12)),
        Frame(send=True, data=PubACK(packet_identifier=1)),
        Frame(send=False, data=b""),
        Frame(send=True, data=Disconnect()),
        ConnectionLoss(),
    ]

    r = SelectReactor()
    f = ReplayClientFactory(r, record)
    e = TCP4ClientEndpoint(r, host, port)
    e.connect(f)
    r.run()

    return Result("qos2_wrong_confirm", f.success, f.reason, f.client_transcript)


def test_qos1_send_wrong_confirm(host, port):
    record = [
        Frame(send=True, data=Connect(client_id="test_wrong_confirm_qos1", flags=ConnectFlags(clean_session=True))),
        Frame(send=False, data=ConnACK(session_present=False, return_code=0)),
        Frame(send=True, data=Subscribe(packet_identifier=1234, topic_requests=[SubscriptionTopicRequest("foo", 2)])),
        Frame(send=False, data=SubACK(packet_identifier=1234, return_codes=[2])),
        Frame(send=True,
              data=Publish(duplicate=False,
                           qos_level=1,
                           topic_name="foo",
                           payload=b"abc",
                           retain=False,
                           packet_identifier=12)),
        Frame(send=False,
              data=[
                  PubACK(packet_identifier=12),
                  Publish(duplicate=False,
                          qos_level=1,
                          topic_name="foo",
                          payload=b"abc",
                          retain=False,
                          packet_identifier=1)
              ]),
        # We send a pubrel to the packet_id expecting a puback
        Frame(send=True, data=PubREL(packet_identifier=1)),
        # ..aaaaand we get a pubcomp back (even though mosquitto warns).
        Frame(send=False, data=PubCOMP(packet_identifier=1)),
        Frame(send=True, data=Disconnect()),
        ConnectionLoss(),
    ]

    r = SelectReactor()
    f = ReplayClientFactory(r, record)
    e = TCP4ClientEndpoint(r, host, port)
    e.connect(f)
    r.run()

    return Result("qos1_wrong_confirm", f.success, f.reason, f.client_transcript)



#####################################################################################
#
#  Copyright (c) Crossbar.io Technologies GmbH
#  SPDX-License-Identifier: EUPL-1.2
#
#####################################################################################

import os
from twisted.trial.unittest import TestCase

from werkzeug.routing import Map, Rule
from jinja2 import Environment, FileSystemLoader
from jinja2.environment import Template

from crossbar.webservice.wap import WapResource


class WapTestCase(TestCase):
    """
    Tests for :class:`crossbar.webservice.wap.WapResource`.
    """

    _WAP1 = {
        "type":
        "wap",
        "templates":
        "../templates",
        "sandbox":
        True,
        "routes": [{
            "path": "/greeting/<name>",
            "method": "GET",
            "call": "com.example.greeting",
            "render": "greeting.html"
        }, {
            "path": "/product/<int:product_id>/<report>/<int:year>/<int:month>",
            "method": "GET",
            "call": "com.example.get_product_report",
            "render": "product_report.html"
        }],
        "wamp": {
            "realm": "realm1",
            "authrole": "anonymous"
        }
    }

    def setUp(self):
        self._templates_dir = os.path.join(os.path.dirname(__file__), 'templates')
        self._jinja_env = Environment(loader=FileSystemLoader(self._templates_dir), autoescape=True)

    def test_map_adapter(self):
        # https://werkzeug.palletsprojects.com/en/2.1.x/routing/#werkzeug.routing.MapAdapter.match
        test_map = Map()
        url = '/reports/product/<int:product_id>/<report>/<int:year>/<int:month>'
        endpoint = 'endpoint1'
        rule = Rule(url, methods=['GET'], endpoint=endpoint)
        test_map.add(rule)
        test_adapter = test_map.bind('localhost', '/')

        test_url = '/reports/product/123/total/2016/12'
        test_data = {'product_id': 123, 'report': 'total', 'year': 2016, 'month': 12}

        _endpoint, _kwargs = test_adapter.match(test_url, method='GET', query_args={})

        self.assertEqual(_endpoint, endpoint)
        self.assertEqual(_kwargs, test_data)

    def test_map_adapter_factory(self):

        map_adapter = WapResource._create_map_adapter(self._jinja_env, self._WAP1, 'localhost', 'reports')

        test_url = '/reports/product/123/total/2016/12'
        test_data = {'product_id': 123, 'report': 'total', 'year': 2016, 'month': 12}
        _endpoint, _kwargs = map_adapter.match(test_url, method='GET', query_args={})

        # ('com.example.get_product_report', <Template 'product_report.html'>) != 'localhost'

        self.assertEqual(_endpoint[0], 'com.example.get_product_report')
        self.assertIsInstance(_endpoint[1], Template)
        self.assertEqual(_kwargs, test_data)



##############################################################################
#
#                        Crossbar.io
#     Copyright (C) Crossbar.io Technologies GmbH. All rights reserved.
#
##############################################################################

import time

import psutil

from crossbar.edge.worker.monitor._base import Monitor
from txaio import perf_counter_ns

__all__ = ('SelfMonitor', )


class SelfMonitor(Monitor):
    """
    Monitor the load induced by the monitoring (native worker) process itself.
    """

    ID = u'self'

    def __init__(self, config=None):
        Monitor.__init__(self, config)
        self._process = psutil.Process()

    def poll(self, sensors=[]):
        """
        Measure current stats value and return new stats.
        """
        hdata = Monitor.poll(self)

        start = perf_counter_ns()

        hdata['io_counters'] = self._process.io_counters()._asdict()
        hdata['cpu_times'] = self._process.cpu_times()._asdict()
        a = round(((self._process.cpu_times().user + self._process.cpu_times().system) * 100) /
                  (time.time() - self._process.create_time()), 3)
        hdata['percent'] = a

        for sensor in sensors:
            hdata[sensor.ID] = sensor._elapsed

        hdata[u'elapsed'] = perf_counter_ns() - start

        self._last_value = hdata

        return hdata



##############################################################################
#
#                        Crossbar.io
#     Copyright (C) Crossbar.io Technologies GmbH. All rights reserved.
#
##############################################################################

from autobahn import wamp

__all__ = ('MarketPlace', )


class MarketPlace(object):
    """
    XBR Market Maker.
    """
    @wamp.register(None)
    def publish(self, publisher_id, api_id, service_id, price=None, details=None):
        """
        Publish a data service implementing a data service API to the market place.

        :param api_id:
        :param service_id:
        :param price:
        :param details:
        :return:
        """
        raise NotImplementedError()



#####################################################################################
#
#  Copyright (c) Crossbar.io Technologies GmbH
#  SPDX-License-Identifier: EUPL-1.2
#
#####################################################################################

from urllib.parse import urljoin

from twisted.internet.defer import inlineCallbacks, returnValue
from twisted.web.http_headers import Headers

from autobahn.twisted.wamp import ApplicationSession


class RESTCallee(ApplicationSession):
    def __init__(self, *args, **kwargs):
        self._webtransport = kwargs.pop("webTransport", None)

        if not self._webtransport:
            import treq
            self._webtransport = treq

        super(RESTCallee, self).__init__(*args, **kwargs)

    @inlineCallbacks
    def onJoin(self, details):
        assert "baseurl" in self.config.extra
        assert "procedure" in self.config.extra

        baseURL = self.config.extra["baseurl"]
        procedure = self.config.extra["procedure"]

        @inlineCallbacks
        def on_call(method=None, url=None, body="", headers={}, params={}):

            newURL = urljoin(baseURL, url)

            params = {x.encode('utf8'): y.encode('utf8') for x, y in params.items()}

            res = yield self._webtransport.request(method,
                                                   newURL,
                                                   data=body.encode('utf8'),
                                                   headers=Headers(headers),
                                                   params=params)
            content = yield self._webtransport.text_content(res)

            headers = {
                x.decode('utf8'): [z.decode('utf8') for z in y]
                for x, y in dict(res.headers.getAllRawHeaders()).items()
            }

            resp = {"code": res.code, "content": content, "headers": headers}

            returnValue(resp)

        yield self.register(on_call, procedure)



#####################################################################################
#
#  Copyright (c) Crossbar.io Technologies GmbH
#  SPDX-License-Identifier: EUPL-1.2
#
#####################################################################################
"""
Interfaces used internally inside Crossbar.io to abstract/decouple different software components.

.. note::

    These interfaces are solely for internal use and not exposed or supposed to be used
    by users or from outside.
"""

import abc
from typing import Union, Dict, List, Any, Optional, Tuple

from autobahn.wamp import ISession
from autobahn.wamp.types import Accept, Deny, HelloDetails, Challenge, CloseDetails, SessionDetails
from autobahn.wamp.message import Publish

from crossbar.router.observation import UriObservationMap

__all__ = (
    'IRealmContainer',
    'IPendingAuth',
    'IRealmStore',
)


class IRealmContainer(abc.ABC):
    """
    Interface to containers of routing realms the authentication system can query
    about the existence of realms and roles during authentication.
    """
    @abc.abstractmethod
    def has_realm(self, realm: str) -> bool:
        """
        Check if a route to a realm with the given name is currently running.

        :param realm: Realm name (the WAMP name, _not_ the run-time object ID).
        :returns: True if a route to the realm (for any role) exists.
        """

    @abc.abstractmethod
    def has_role(self, realm: str, authrole: str) -> bool:
        """
        Check if a role with the given name is currently running in the given realm.

        :param realm: WAMP realm (the WAMP name, _not_ the run-time object ID).
        :param authrole: WAMP authentication role (the WAMP URI, _not_ the run-time object ID).
        :returns: True if a route to the realm for the role exists.
        """

    @abc.abstractmethod
    def get_service_session(self, realm: str, role: str) -> ISession:
        """
        Returns a service session on the given realm using the given role.
        Service sessions are used for:

        * access dynamic authenticators (see :method:`crossbar.router.auth.pending.PendingAuth._init_dynamic_authenticator`)
        * access the WAMP meta API for the realm (see :method:`crossbar.worker.router.RouterController.start_router_realm`)
        * forward to/from WAMP for the HTTP bridge (see :class:`crossbar.webservice.rest.RouterWebServiceRestPublisher`, :class:`crossbar.webservice.rest.RouterWebServiceRestCaller`, :class:`crossbar.webservice.rest.RouterWebServiceWebhook`)

        :param realm: WAMP realm name.
        :param role: WAMP authentication role name.
        :returns: A service session joined on the given realm and role.
        """


class IPendingAuth(abc.ABC):
    """
    Interface to pending WAMP authentications.
    """
    @abc.abstractmethod
    def hello(self, realm: str, details: HelloDetails) -> Union[Accept, Deny, Challenge]:
        """
        When a HELLO message is received, this gets called to open the pending authentication.

        :param realm: The realm to client wishes to join (if the client did announce a realm).
        :param details: The details of the client provided for HELLO.
        :returns: Either return a challenge, or immediately accept or deny session.
        """

    @abc.abstractmethod
    def authenticate(self, signature: str) -> Union[Accept, Deny]:
        """
        The client has answered with a WAMP AUTHENTICATE message. Verify the message and accept or deny.

        :param signature: Signature over the challenge as received from the authenticating session.
        :returns: Either accept or deny the session.
        """


class IRealmStore(abc.ABC):
    """
    Realm store interface common to transient and persistent store implementations.
    """
    @property
    @abc.abstractmethod
    def type(self) -> str:
        """

        :return: Return type of realm store, e.g. ``"memory"`` or ``"cfxdb"``.
        """

    @property
    @abc.abstractmethod
    def is_running(self) -> bool:
        """

        :return: True if this realm store is currently running.
        """

    @abc.abstractmethod
    def start(self):
        """
        Initialize and start this realm store.

        .. note::

            This procedure may or may not run asynchronously
            depending on store type.
        """

    @abc.abstractmethod
    def stop(self):
        """
        Stop this realm store and retain all data.

        .. note::

            This procedure may or may not run asynchronously
            depending on store type.
        """

    @abc.abstractmethod
    def store_session_joined(self, session: ISession, details: SessionDetails):
        """

        :param session: Session that has joined a realm.
        :param details: Session details of the joined session.
        """

    @abc.abstractmethod
    def store_session_left(self, session: ISession, details: CloseDetails):
        """

        :param session: Session that has left a realm it was previously joined on.
        :param details: Session close details.
        """

    @abc.abstractmethod
    def get_session_by_session_id(self, session_id: int, joined_at: Optional[int] = None) -> Optional[Dict[str, Any]]:
        """
        Get session information by WAMP session ID. If there is no session stored, return ``None``.

        .. note::

            In the rare event there is more than one session for the same WAMP session ID
            (which may happen as the WAMP session ID is effectively 56 bit), the procedure
            will return the information for the session with the most recent ``joined_at``
            timestamp.

        :param session_id: The WAMP session ID of the session for which to return information.
        :param joined_at: If there are more than one session for the given ``session_id``,
            only return the session with the given joining timestamp (or newer).
        :return: A set of information for the stored session.
        """

    @abc.abstractmethod
    def get_sessions_by_authid(self, authid: str) -> Optional[List[Tuple[str, int]]]:
        """
        Return session IDs and session joined timestamps given a WAMP authid, ordered by
        timestamps in reverse chronological order.

        :param authid: The WAMP authid to retrieve sessions for.
        :return: List of pairs ``(session_id, joined_at)`` or ``None`` if there are no
            sessions stored for the given ``authid``.
        """

    @abc.abstractmethod
    def attach_subscription_map(self, subscription_map: UriObservationMap):
        """

        :param subscription_map:
        :return:
        """

    @abc.abstractmethod
    def store_event(self, session: ISession, publication_id: int, publish: Publish):
        """
        Store event to event history.

        :param session: The publishing session.
        :param publication_id: The WAMP publication ID under which the publish-action happens.
        :param publish: The WAMP publish message.
        """

    @abc.abstractmethod
    def store_event_history(self, publication_id: int, subscription_id: int, receiver: ISession):
        """
        Store publication history for subscription.

        :param publication_id: The ID of the event publication to be persisted.
        :param subscription_id: The ID of the subscription the event (identified by the publication ID),
            was published to, because the event's topic matched the subscription.
        :param receiver: The receiving session.
        """

    @abc.abstractmethod
    def get_events(self, subscription_id: int, limit: Optional[int] = None):
        """
        Retrieve given number of last events for a given subscription.

        If no events are yet stored, an empty list ``[]`` is returned.
        If no history is maintained at all for the given subscription, ``None`` is returned.

        This procedure is called by the service session of Crossbar.io and
        exposed under the WAMP meta API procedure ``wamp.subscription.get_events``.

        :param subscription_id: The ID of the subscription to retrieve events for.
        :param limit: Limit number of events returned.
        :return: List of events: at most ``limit`` events in reverse chronological order.
        """

    @abc.abstractmethod
    def get_event_history(self,
                          subscription_id: int,
                          from_ts: int,
                          until_ts: int,
                          reverse: Optional[bool] = None,
                          limit: Optional[int] = None) -> Optional[List[Dict[str, Any]]]:
        """
        Retrieve event history for time range for a given subscription.

        If no history is maintained for the given subscription, None is returned.

        :param subscription_id: The ID of the subscription to retrieve events for.
        :param from_ts: Filter events from this date (epoch time in ns).
        :param until_ts: Filter events until before this date (epoch time in ns).
        :param reverse:
        :param limit:
        """

    @abc.abstractmethod
    def maybe_queue_call(self, session: ISession, call, registration, authorization):
        """

        :param session:
        :param call:
        :param registration:
        :param authorization:
        :return:
        """

    @abc.abstractmethod
    def get_queued_call(self, registration):
        """

        :param registration:
        :return:
        """

    @abc.abstractmethod
    def pop_queued_call(self, registration):
        """

        :param registration:
        :return:
        """



###############################################################################
#
# Crossbar.io Shell
# Copyright (c) Crossbar.io Technologies GmbH. Licensed under EUPLv1.2.
#
###############################################################################

from .BaseType import BaseType
from .Enum import Enum
from .EnumVal import EnumVal
from .Field import Field
from .KeyValue import KeyValue
from .Object import Object
from .RPCCall import RPCCall
from .Schema import Schema
from .Service import Service
from .Type import Type



import time
import sys
import os

def test_import_1():
    started = time.monotonic_ns()
    import autobahn
    ended = time.monotonic_ns()
    return ended - started

def test_import_2():
    started = time.monotonic_ns()
    from autobahn import xbr
    ended = time.monotonic_ns()
    return ended - started

def test_import_3():
    started = time.monotonic_ns()
    from crossbar import personality as standalone
    ended = time.monotonic_ns()
    return ended - started

def test_import_4():
    started = time.monotonic_ns()
    from crossbar import personality as standalone
    from crossbar import edge
    from crossbar import network
    from crossbar import master
    from crossbar.node.main import main
    ended = time.monotonic_ns()
    return ended - started

def test_import_5():
    started = time.monotonic_ns()
    import zlmdb
    ended = time.monotonic_ns()
    return ended - started

def test_import_6():
    started = time.monotonic_ns()
    import cfxdb
    ended = time.monotonic_ns()
    return ended - started

def test_import_7():
    started = time.monotonic_ns()
    import autobahn
    import cbor2
    import flatbuffers
    import numpy
    import multihash
    import txaio
    import click
    import web3
    import zlmdb
    ended = time.monotonic_ns()
    return ended - started

tests = {
    '1': (test_import_1, 'only autobahn'),
    '2': (test_import_2, 'xbr from autobahn'),
    '3': (test_import_3, 'crossbar oss'),
    '4': (test_import_4, 'crossbar full'),
    '5': (test_import_5, 'only zlmdb'),
    '6': (test_import_6, 'cfxdb'),
    '7': (test_import_7, 'all cfxdb deps'),
}

test = None
if len(sys.argv) > 1:
    test, test_desc = tests.get(sys.argv[1], (None, None))

if test:
    dur = test()
    print('test {} ("{}") ran in {} seconds'.format(sys.argv[1], test_desc, dur / 10**9))
    print()
else:
    for i in range(len(tests)):
        cmd = ' '.join([sys.executable, __file__, str(i + 1)])
        os.system(cmd)



###############################################################################
#
# Crossbar.io Master
# Copyright (c) Crossbar.io Technologies GmbH. Licensed under EUPLv1.2.
#
###############################################################################

import uuid
from typing import Optional, List
from pprint import pformat

import numpy as np
from sortedcontainers import SortedDict

from autobahn import wamp
from autobahn.wamp.exception import ApplicationError
from autobahn.wamp.types import CallDetails, PublishOptions, RegisterOptions

from crossbar.common import checkconfig
from crossbar._util import hl, hlid, hltype, hlval
from cfxdb.mrealm import RouterCluster, RouterClusterNodeMembership, RouterWorkerGroupClusterPlacement
from cfxdb.mrealm import cluster, RouterWorkerGroup, WorkerGroupStatus

import txaio

txaio.use_twisted()
from txaio import time_ns, sleep, make_logger  # noqa
from twisted.internet.defer import inlineCallbacks
from twisted.internet.task import LoopingCall


class RouterClusterMonitor(object):
    """
    Background monitor running periodically in the master node to monitor, check and apply
    necessary actions for router clusters.

    The monitor is started when a router cluster is started.
    """
    log = make_logger()

    def __init__(self, manager, routercluster_oid, interval=10.):
        self._manager = manager
        self._routercluster_oid = routercluster_oid
        self._interval = interval

        self._loop = None
        self._check_and_apply_in_progress = False

    @property
    def is_started(self):
        """

        :return:
        """
        return self._loop is not None and self._loop.running

    def start(self):
        """

        :return:
        """
        assert self._loop is None

        self._loop = LoopingCall(self._check_and_apply)
        self._loop.start(self._interval)

    def stop(self):
        """

        :return:
        """
        assert self._loop is not None

        self._loop.stop()
        self._loop = None
        self._check_and_apply_in_progress = False

    @inlineCallbacks
    def _check_and_apply(self):
        if self._check_and_apply_in_progress:
            # we prohibit running the iteration multiple times concurrently. this might
            # happen when the iteration takes longer than the interval the monitor is set to
            self.log.warn(
                '{func} {action} for routercluster {routercluster} skipped! check & apply already in progress.',
                action=hl('check & apply run skipped', color='red', bold=True),
                func=hltype(self._check_and_apply),
                routercluster=hlid(self._routercluster_oid))
            return
        else:
            self.log.info('{func} {action} for routercluster {routercluster} ..',
                          action=hl('check & apply run started', color='green', bold=True),
                          func=hltype(self._check_and_apply),
                          routercluster=hlid(self._routercluster_oid))
            self._check_and_apply_in_progress = True

        is_running_completely = True
        try:
            # get all active (non-standby) nodes added to the routercluster
            active_memberships = []
            with self._manager.db.begin(write=False) as txn:
                # the routercluster itself
                routercluster = self._manager.schema.routerclusters[txn, self._routercluster_oid]

                if routercluster.status in [cluster.STATUS_STARTING, cluster.STATUS_RUNNING]:
                    # the node memberships in the routercluster
                    active_memberships = [
                        m for m in self._manager.schema.routercluster_node_memberships.select(
                            txn, from_key=(routercluster.oid, uuid.UUID(bytes=b'\0' * 16)), return_keys=False)
                    ]

            for membership in active_memberships:
                node_oid = str(membership.node_oid)

                # node run-time information, as maintained here in our master view of the external world
                node = self._manager._session.nodes.get(node_oid, None)

                if node and node.status == 'online':
                    self.log.info('{func} Ok, router cluster node {node_oid} is running!',
                                  func=hltype(self._check_and_apply),
                                  node_oid=hlid(node_oid))

                    # FIXME: check all workers we expect for data planes associated with this router cluster are running

                else:
                    self.log.warn('{func} Router cluster node {node_oid} not running [status={status}]',
                                  func=hltype(self._check_and_apply),
                                  node_oid=hlid(node_oid),
                                  status=hl(node.status if node else 'offline'))
                    is_running_completely = False

            if routercluster.status in [cluster.STATUS_STARTING] and is_running_completely:
                with self._manager.db.begin(write=True) as txn:
                    routercluster = self._manager.schema.routerclusters[txn, self._routercluster_oid]
                    routercluster.status = cluster.STATUS_RUNNING
                    routercluster.changed = time_ns()
                    self._manager.schema.routerclusters[txn, routercluster.oid] = routercluster

                routercluster_started = {
                    'oid': str(routercluster.oid),
                    'status': cluster.STATUS_BY_CODE[routercluster.status],
                    'changed': routercluster.changed,
                }
                yield self._manager._session.publish('{}.on_routercluster_started'.format(self._manager._prefix),
                                                     routercluster_started,
                                                     options=self._manager._PUBOPTS)
        except:
            self.log.failure()

        if is_running_completely:
            color = 'green'
            action = 'check & apply run completed successfully'
        else:
            color = 'red'
            action = 'check & apply run finished with problems left'

        self._check_and_apply_in_progress = False
        self.log.info('{func} {action} for routercluster {routercluster}!',
                      action=hl(action, color=color, bold=True),
                      func=hltype(self._check_and_apply),
                      routercluster=hlid(self._routercluster_oid))


class RouterClusterManager(object):
    """
    Manages Router clusters, which runs Crossbar.io Web transport listening
    endpoints on many (frontend) workers over many nodes using applying
    a shared, common transport definition, such as regarding the Web services
    configured on URL paths of the Web transport.

    - routercluster
      - routercluster nodes
      - routercluster workergroup
    """
    log = make_logger()

    # publication options for management API events
    _PUBOPTS = PublishOptions(acknowledge=True)

    def __init__(self, session, globaldb, globalschema, db, schema, reactor=None):
        """

        :param session: Backend of user created management realms.
        :type session: :class:`crossbar.master.mrealm.controller.MrealmController`

        :param globaldb: Global database handle.
        :type globaldb: :class:`zlmdb.Database`

        :param globalschema: Global database schema.
        :type globalschema: :class:`cfxdb.globalschema.GlobalSchema`

        :param db: Management realm database handle.
        :type db: :class:`zlmdb.Database`

        :param schema: Management realm database schema.
        :type schema: :class:`cfxdb.mrealmschema.MrealmSchema`
        """
        self._session = session

        # ContainerController
        self._worker = session.config.controller

        # node personality
        self._personality = self._worker.personality

        # twisted reactor
        self._reactor = self._worker._reactor
        # if not reactor:
        #     from twisted.internet import reactor
        # self._reactor = reactor

        # global database handle & schema

        # global (node level) database handle & schema
        self.gdb = globaldb
        self.gschema = globalschema

        # mrealm database handle & schema
        self.db = db
        self.schema = schema

        # the management realm OID this routercluster manager operates for
        self._mrealm_oid = session._mrealm_oid

        # URI prefix for WAMP procedures/topics, filled when started
        self._prefix = None

        # filled when started
        self._started = None

        # router cluster monitors, containing a map, for every routercluster in state STARTING or RUNNING
        # with objects of class RouterClusterMonitor
        self._monitors = {}

    @inlineCallbacks
    def start(self, prefix):
        """
        Start this router cluster manager, including all monitors of router clusters defined.

        :return:
        """
        assert self._started is None, 'cannot start router cluster manager - already running!'
        assert self._prefix is None

        self._started = time_ns()

        # crossbarfabriccenter.mrealm.routercluster
        self._prefix = prefix[:-1] if prefix.endswith('.') else prefix

        # register management procedures
        regs = yield self._session.register(self,
                                            prefix='{}.'.format(self._prefix),
                                            options=RegisterOptions(details_arg='details'))
        procs = [reg.procedure for reg in regs]
        self.log.info(
            'Router cluster manager registered {api} management procedures using prefix "{prefix}" [{func}]:\n\n{procs}\n',
            api=hl('Router cluster manager API', color='green', bold=True),
            func=hltype(self.start),
            prefix=hlval(self._prefix),
            procs=hl(pformat(procs), color='white', bold=True))

        # start all router cluster monitors
        cnt_started = 0
        cnt_skipped = 0
        dl = []
        with self.db.begin() as txn:
            routercluster_oids = self.schema.routerclusters.select(txn, return_values=False)
            for routercluster_oid in routercluster_oids:
                routercluster = self.schema.routerclusters[txn, routercluster_oid]
                if routercluster.status in [cluster.STATUS_STARTING, cluster.STATUS_RUNNING]:
                    assert routercluster_oid not in self._monitors
                    monitor = RouterClusterMonitor(self, routercluster_oid)
                    dl.append(monitor.start())
                    self._monitors[routercluster_oid] = monitor
                    cnt_started += 1
                    self.log.info(
                        'Router cluster monitor started for router cluster {routercluster_oid} in status {status} [{func}]',
                        func=hltype(self.start),
                        routercluster_oid=hlid(routercluster_oid),
                        status=hlval(routercluster.status))
                else:
                    cnt_skipped += 1
                    self.log.info(
                        'Router cluster monitor skipped for router cluster {routercluster_oid} in status {status} [{func}]',
                        func=hltype(self.start),
                        routercluster_oid=hlid(routercluster_oid),
                        status=hlval(routercluster.status))
        self.log.info(
            'Router cluster manager has started monitors for {cnt_started} clusters ({cnt_skipped} skipped) [{func}]',
            cnt_started=hlval(cnt_started),
            cnt_skipped=hlval(cnt_skipped),
            func=hltype(self.start))

        self.log.info('Router cluster manager for management realm {mrealm_oid} ready [{func}]',
                      mrealm_oid=hlid(self._mrealm_oid),
                      func=hltype(self.start))

        # return txaio.gather(dl)

    def stop(self):
        """
        Stop the currently running router cluster manager. This will stop all monitors for router clusters.

        :return:
        """
        assert self._started > 0, 'cannot stop router cluster manager - currently not running!'

        # stop all router cluster monitors ..
        dl = []
        for routercluster_oid, routercluster_monitor in self._monitors.items():
            dl.append(routercluster_monitor.stop())
            del self._monitors[routercluster_oid]
        self._started = None
        self.log.info(
            'Ok, router cluster manager for management realm {mrealm_oid} stopped ({cnt_stopped} monitors stopped) [{func}]',
            mrealm_oid=hlid(self._mrealm_oid),
            cnt_stopped=len(dl),
            func=hltype(self.start))

        # return txaio.gather(dl)

    @wamp.register(None, check_types=True)
    def list_routerclusters(self,
                            return_names: Optional[bool] = None,
                            details: Optional[CallDetails] = None) -> List[str]:
        """
        Returns list of router clusters defined. Detail information for a router cluster
        can be retrieved using :meth:`crossbar.master.cluster.routercluster.RouterClusterManager.get_routercluster`.

        :param return_names: Return router clusters names instead of object IDs.

        :return: List of router clusters object IDs or names. For example:

            .. code-block:: json

                [
                    "634e0725-df03-4daf-becd-1de60dd2b0b3",
                    "7dc55a4e-e52a-4bea-a8b4-daf869cc417f"
                ]

            or with ``return_names``  set:

            .. code-block:: json

                [
                    "cluster1"
                ]
        """
        self.log.info('{func}(details={details})', func=hltype(self.list_routerclusters), details=details)

        with self.db.begin() as txn:
            if return_names:
                routerclusters = self.schema.routerclusters.select(txn, return_keys=False)
                if routerclusters:
                    return sorted([routercluster.name for routercluster in routerclusters])
                else:
                    return []
            else:
                routercluster_oids = self.schema.routerclusters.select(txn, return_values=False)
                if routercluster_oids:
                    # we now have a list of uuid.UUID objects: convert to strings
                    return [str(oid) for oid in routercluster_oids]
                else:
                    return []

    @wamp.register(None, check_types=True)
    def get_routercluster(self, routercluster_oid: str, details: Optional[CallDetails] = None) -> dict:
        """
        Return configuration and run-time status information for a router cluster (by object ID).

        :param routercluster_oid: Object ID of the router cluster to return.

        :return: Router cluster definition. For example, initially, after a router cluster has been created:

            .. code-block:: json

                {
                    "changed": 1598273658338443875,
                    "description": null,
                    "label": null,
                    "name": "cluster2",
                    "oid": "634e0725-df03-4daf-becd-1de60dd2b0b3",
                    "status": "STOPPED",
                    "tags": null
                }
        """
        self.log.info('{func}(routercluster_oid={routercluster_oid}, details={details})',
                      func=hltype(self.get_routercluster),
                      routercluster_oid=hlid(routercluster_oid),
                      details=details)

        try:
            routercluster_oid_ = uuid.UUID(routercluster_oid)
        except Exception as e:
            raise ApplicationError('wamp.error.invalid_argument', 'invalid routercluster_oid: {}'.format(str(e)))

        with self.db.begin() as txn:
            routercluster = self.schema.routerclusters[txn, routercluster_oid_]

        if routercluster:
            return routercluster.marshal()
        else:
            raise ApplicationError('crossbar.error.no_such_object',
                                   'no routercluster with oid {}'.format(routercluster_oid_))

    @wamp.register(None, check_types=True)
    def get_routercluster_by_name(self, routercluster_name: str, details: Optional[CallDetails] = None) -> dict:
        """
        Return configuration and run-time status information for a router cluster (by name).

        See also the corresponding procedure :meth:`crossbar.master.cluster.routercluster.RouterClusterManager.get_routercluster`
        which returns the same information, given and object ID rather than name.

        :param routercluster_name: Name of the router cluster to return.

        :return: Router cluster definition.
        """
        self.log.info('{func}(routercluster_name="{routercluster_name}", details={details})',
                      func=hltype(self.get_routercluster_by_name),
                      routercluster_name=hlid(routercluster_name),
                      details=details)

        with self.db.begin() as txn:
            routercluster_oid = self.schema.idx_routerclusters_by_name[txn, routercluster_name]
            if not routercluster_oid:
                raise ApplicationError('crossbar.error.no_such_object',
                                       'no routercluster named {}'.format(routercluster_name))

            routercluster = self.schema.routerclusters[txn, routercluster_oid]
            assert routercluster

        return routercluster.marshal()

    @wamp.register(None, check_types=True)
    async def create_routercluster(self, routercluster: dict, details: Optional[CallDetails] = None) -> dict:
        """
        Create a new router cluster definition.

        :procedure: ``crossbarfabriccenter.mrealm.cluster.create_routercluster`` URI of WAMP procedure to call.
        :event: ``crossbarfabriccenter.mrealm.cluster.on_routercluster_created`` WAMP event published once the
            router cluster has been created.
        :error: ``wamp.error.invalid_configuration`` WAMP error returned when the router cluster
            configuration provided has a problem.
        :error: ``wamp.error.not_authorized`` WAMP error returned when the user is currently not allowed
            to created (another) router cluster.
        :error: ``crossbar.error.already_exists`` WAMP error returned when a router cluster named as contained
            in the configuration already exists.

        :param routercluster: Router cluster settings. For example:

            .. code-block:: json

                {
                    "name": "cluster5"
                }

        :return: Router cluster creation information. For example:

            .. code-block:: json

                {
                    "changed": 1598379288123799334,
                    "description": null,
                    "label": null,
                    "name": "cluster5",
                    "oid": "3eccb1fd-251b-4eda-bee9-06b3d24b1c5e",
                    "owner_oid": "f1c62815-56b2-484f-bb5a-a66a788c2aff",
                    "status": "STOPPED",
                    "tags": null
                }
        """
        self.log.info('{func}(routercluster="{routercluster}", details={details})',
                      func=hltype(self.create_routercluster),
                      routercluster=routercluster,
                      details=details)

        try:
            obj = RouterCluster.parse(routercluster)
        except Exception as e:
            raise ApplicationError('wamp.error.invalid_configuration',
                                   'could not parse router cluster configuration ({})'.format(e))

        if obj.name is None:
            raise ApplicationError('wamp.error.invalid_configuration',
                                   'missing "name" in router cluster configuration')
        else:
            if not checkconfig._CONFIG_ITEM_ID_PAT.match(obj.name):
                raise ApplicationError(
                    'wamp.error.invalid_configuration',
                    'invalid name "{}" in router cluster configuration (must match {})'.format(
                        obj.name, checkconfig._CONFIG_ITEM_ID_PAT_STR))

        obj.oid = uuid.uuid4()
        obj.status = cluster.STATUS_BY_NAME['STOPPED']
        obj.changed = np.datetime64(time_ns(), 'ns')

        if details and details.caller_authid:
            with self.gdb.begin() as txn:
                caller_oid = self.gschema.idx_users_by_email[txn, details.caller_authid]
                if not caller_oid:
                    raise ApplicationError('wamp.error.no_such_principal',
                                           'no user found for authid "{}"'.format(details.caller_authid))
            obj.owner_oid = caller_oid
        else:
            raise ApplicationError('wamp.error.no_such_principal', 'cannot map user - no caller authid available')

        with self.db.begin(write=True) as txn:
            if self.schema.idx_routerclusters_by_name[txn, obj.name]:
                raise ApplicationError('crossbar.error.already_exists',
                                       'duplicate name "{}" in router cluster configuration'.format(obj.name))

            self.schema.routerclusters[txn, obj.oid] = obj

        self.log.info('new RouterCluster object stored in database:\n{obj}', obj=obj)

        res_obj = obj.marshal()

        await self._session.publish('{}.on_routercluster_created'.format(self._prefix), res_obj, options=self._PUBOPTS)

        self.log.info('Management API event <on_routercluster_created> published:\n{res_obj}', res_obj=res_obj)

        return res_obj

    @wamp.register(None, check_types=True)
    async def delete_routercluster(self, routercluster_oid: str, details: Optional[CallDetails] = None) -> dict:
        """
        Delete an existing router cluster definition. The router cluster must be in status ``"STOPPED"``.

        :procedure: ``crossbarfabriccenter.routercluster.delete_routercluster`` URI of WAMP procedure to call.
        :event: ``crossbarfabriccenter.routercluster.on_routercluster_deleted`` WAMP event published once the router cluster has been deleted.
        :error: ``wamp.error.invalid_argument`` WAMP error returned when ``routercluster_oid`` was invalid.
        :error: ``crossbar.error.no_such_object`` WAMP error returned when ``routercluster_oid`` was not found.
        :error: ``crossbar.error.not_stopped`` WAMP error returned when router cluster is not in status ``STOPPED``.

        :param routercluster_oid: OID of the router cluster to delete

        :returns: Deleted router cluster, for example:

            .. code-block:: json

                {
                    "changed": 1598380973225053489,
                    "description": null,
                    "label": null,
                    "name": "cluster5",
                    "oid": "3eccb1fd-251b-4eda-bee9-06b3d24b1c5e",
                    "owner_oid": "f1c62815-56b2-484f-bb5a-a66a788c2aff",
                    "status": "STOPPED",
                    "tags": null
                }
        """
        self.log.info('{func}.delete_routercluster(routercluster_oid={routercluster_oid}, details={details})',
                      func=hltype(self.delete_routercluster),
                      routercluster_oid=hlid(routercluster_oid),
                      details=details)

        try:
            oid = uuid.UUID(routercluster_oid)
        except Exception as e:
            raise ApplicationError('wamp.error.invalid_argument', 'invalid oid "{}"'.format(str(e)))

        if details and details.caller_authid:
            with self.gdb.begin() as txn:
                caller_oid = self.gschema.idx_users_by_email[txn, details.caller_authid]
                if not caller_oid:
                    raise ApplicationError('wamp.error.no_such_principal',
                                           'no user found for authid "{}"'.format(details.caller_authid))
        else:
            raise ApplicationError('wamp.error.no_such_principal', 'cannot map user - no caller authid available')

        with self.db.begin(write=True) as txn:
            cluster_obj = self.schema.routerclusters[txn, oid]
            if cluster_obj:
                if cluster_obj.owner_oid != caller_oid:
                    raise ApplicationError('wamp.error.not_authorized',
                                           'only owner is allowed to delete router cluster')
                if cluster_obj.status != cluster.STATUS_STOPPED:
                    raise ApplicationError('crossbar.error.not_stopped')
                del self.schema.routerclusters[txn, oid]
            else:
                raise ApplicationError('crossbar.error.no_such_object', 'no object with oid {} found'.format(oid))

        cluster_obj.changed = time_ns()
        self.log.info('RouterCluster object deleted from database:\n{cluster}', cluster=cluster_obj)

        res_obj = cluster_obj.marshal()
        await self._session.publish('{}.on_routercluster_deleted'.format(self._prefix), res_obj, options=self._PUBOPTS)
        return res_obj

    @wamp.register(None, check_types=True)
    async def start_routercluster(self, routercluster_oid: str, details: Optional[CallDetails] = None) -> dict:
        """
        Start a router cluster

        :param routercluster_oid: Object ID of router cluster to start.

        :return: Started router cluster, for example:

            .. code-block:: json

                {
                    "changed": 1598402748823470105,
                    "oid": "ad6cfb53-3712-4683-8b15-f48a6d71d410",
                    "status": "STARTING",
                    "who": {
                        "authid": "superuser",
                        "authrole": "owner",
                        "session": 6761363113437744
                    }
                }
        """
        self.log.info('{func}(routercluster_oid="{routercluster_oid}", details={details})',
                      func=hltype(self.start_routercluster),
                      routercluster_oid=hlid(routercluster_oid),
                      details=details)

        try:
            routercluster_oid_ = uuid.UUID(routercluster_oid)
        except Exception as e:
            raise ApplicationError('wamp.error.invalid_argument', 'invalid oid "{}"'.format(str(e)))

        if details and details.caller_authid:
            with self.gdb.begin() as txn:
                caller_oid = self.gschema.idx_users_by_email[txn, details.caller_authid]
                if not caller_oid:
                    raise ApplicationError('wamp.error.no_such_principal',
                                           'no user found for authid "{}"'.format(details.caller_authid))
        else:
            raise ApplicationError('wamp.error.no_such_principal', 'cannot map user - no caller authid available')

        with self.db.begin(write=True) as txn:
            routercluster = self.schema.routerclusters[txn, routercluster_oid_]
            if not routercluster:
                raise ApplicationError('crossbar.error.no_such_object',
                                       'no routercluster with oid {} found'.format(routercluster_oid_))

            if routercluster.owner_oid != caller_oid:
                raise ApplicationError('wamp.error.not_authorized', 'only owner is allowed to start a router cluster')

            if routercluster.status not in [cluster.STATUS_STOPPED, cluster.STATUS_PAUSED]:
                emsg = 'cannot start routercluster currently in state {}'.format(
                    cluster.STATUS_BY_CODE[routercluster.status])
                raise ApplicationError('crossbar.error.cannot_start', emsg)

            routercluster.status = cluster.STATUS_STARTING
            routercluster.changed = time_ns()

            self.schema.routerclusters[txn, routercluster_oid_] = routercluster

        monitor = RouterClusterMonitor(self, routercluster_oid_)
        monitor.start()
        assert routercluster_oid_ not in self._monitors
        self._monitors[routercluster_oid_] = monitor

        routercluster_starting = {
            'oid': str(routercluster.oid),
            'status': cluster.STATUS_BY_CODE[routercluster.status],
            'changed': routercluster.changed,
            'who': {
                'session': details.caller if details else None,
                'authid': details.caller_authid if details else None,
                'authrole': details.caller_authrole if details else None,
            }
        }
        await self._session.publish('{}.on_routercluster_starting'.format(self._prefix),
                                    routercluster_starting,
                                    options=self._PUBOPTS)

        return routercluster_starting

    @wamp.register(None, check_types=True)
    async def stop_routercluster(self, routercluster_oid: str, details: Optional[CallDetails] = None) -> dict:
        """
        Stop a running router cluster.

        :param routercluster_oid: Object ID of router cluster to stop.

        :return: Stopped router cluster, for example:

            .. code-block:: json

                {
                    "changed": 1598402964397396934,
                    "oid": "ad6cfb53-3712-4683-8b15-f48a6d71d410",
                    "status": "STOPPING",
                    "who": {
                        "authid": "superuser",
                        "authrole": "owner",
                        "session": 8299909547427073
                    }
                }
        """
        self.log.info('{func}(routercluster_oid={routercluster_oid}, details={details})',
                      routercluster_oid=hlid(routercluster_oid),
                      func=hltype(self.stop_routercluster),
                      details=details)

        try:
            routercluster_oid_ = uuid.UUID(routercluster_oid)
        except Exception as e:
            raise ApplicationError('wamp.error.invalid_argument', 'invalid oid "{}"'.format(str(e)))

        with self.db.begin(write=True) as txn:
            routercluster = self.schema.routerclusters[txn, routercluster_oid_]
            if not routercluster:
                raise ApplicationError('crossbar.error.no_such_object',
                                       'no routercluster with oid {} found'.format(routercluster_oid_))

            if routercluster.status not in [cluster.STATUS_STARTING, cluster.STATUS_RUNNING]:
                emsg = 'cannot stop routercluster currently in state {}'.format(
                    cluster.STATUS_BY_CODE[routercluster.status])
                raise ApplicationError('crossbar.error.cannot_start', emsg)

            routercluster.status = cluster.STATUS_STOPPING
            routercluster.changed = time_ns()

            self.schema.routerclusters[txn, routercluster_oid_] = routercluster

        routercluster_stopping = {
            'oid': str(routercluster.oid),
            'status': cluster.STATUS_BY_CODE[routercluster.status],
            'changed': routercluster.changed,
            'who': {
                'session': details.caller if details else None,
                'authid': details.caller_authid if details else None,
                'authrole': details.caller_authrole if details else None,
            }
        }

        await self._session.publish('{}.on_routercluster_stopping'.format(self._prefix),
                                    routercluster_stopping,
                                    options=self._PUBOPTS)

        return routercluster_stopping

    @wamp.register(None, check_types=True)
    def stat_routercluster(self, routercluster_oid: str, details: Optional[CallDetails] = None) -> dict:
        """
        *NOT YET IMPLEMENTED*

        Get current status and statistics for given router cluster.

        :param routercluster_oid: The router cluster to return status and statistics for.

        :return: Current status and statistics for given router cluster.
        """
        self.log.info('{func}(routercluster_oid={routercluster_oid}, details={details})',
                      func=hltype(self.stat_routercluster),
                      routercluster_oid=hlid(routercluster_oid),
                      details=details)

        raise NotImplementedError()

    @wamp.register(None, check_types=True)
    def list_routercluster_nodes(self,
                                 routercluster_oid: str,
                                 return_names: Optional[bool] = None,
                                 filter_by_status: Optional[str] = None,
                                 details: Optional[CallDetails] = None) -> List[str]:
        """
        List nodes currently associated with the given router cluster.

        :param routercluster_oid: The router cluster to list nodes for.
        :param return_names: Return routercluster names instead of  object IDs
        :param filter_by_status: Filter nodes by this status, eg. ``"online"``.

        :return: List of node IDs of nodes associated with the router cluster. For example:

            .. code-block:: json

                [
                    "0afb5897-d8da-433a-9214-ed64e8da50b9",
                    "2f656d47-5251-44bb-a507-6cebc533eb50",
                    "7ddf39c5-6752-4467-9497-3f1758b2ac5e",
                    "879d05f3-e3d3-4bce-894e-a281e4782a0b"
                ]

            or with ``return_names`` set:

            .. code-block:: json

                [
                    "node1",
                    "node2",
                    "node3",
                    "node4"
                ]
        """
        self.log.info(
            '{func}(routercluster_oid={routercluster_oid}, return_names={return_names}, filter_by_status={filter_by_status}, details={details})',
            func=hltype(self.list_routercluster_nodes),
            routercluster_oid=hlid(routercluster_oid),
            return_names=hlval(return_names),
            filter_by_status=hlval(filter_by_status),
            details=details)

        try:
            routercluster_oid_ = uuid.UUID(routercluster_oid)
        except Exception as e:
            raise ApplicationError('wamp.error.invalid_argument', 'invalid oid "{}"'.format(str(e)))

        node_oids = []
        with self.db.begin() as txn:
            routercluster = self.schema.routerclusters[txn, routercluster_oid_]
            if not routercluster:
                raise ApplicationError('crossbar.error.no_such_object',
                                       'no routercluster with oid {} found'.format(routercluster_oid_))
            for _, node_oid in self.schema.routercluster_node_memberships.select(
                    txn, from_key=(routercluster_oid_, uuid.UUID(bytes=b'\0' * 16)), return_values=False):
                node_oids.append(node_oid)

        if filter_by_status:
            node_oids_ = []
            for node_oid in node_oids:
                node = self._session.nodes.get(str(node_oid), None)
                if node and node.status == filter_by_status:
                    node_oids_.append(node_oid)
            node_oids = node_oids_

        if return_names:
            node_authids = []
            with self.gdb.begin() as txn:
                for node_oid in node_oids:
                    node = self.gschema.nodes[txn, node_oid]
                    if node and node.authid:
                        node_authids.append(node.authid)
            res = sorted(node_authids)
        else:
            res = [str(node_oid) for node_oid in node_oids]

        return res

    @wamp.register(None, check_types=True)
    async def add_routercluster_node(self,
                                     routercluster_oid: str,
                                     node_oid: str,
                                     config: Optional[dict] = None,
                                     details: Optional[CallDetails] = None) -> dict:
        """
        Add a node to a router cluster. You can configure the node association for the
        cluster using ``config``:

        - ``hardlimit``: hard limit on node utilization (number of workers run on this node)
        - ``softlimit``: soft limit on node utilization (number of workers run on this node)

        :param routercluster_oid: OID of the router cluster to which to add the node.
        :param node_oid: OID of the node to add to the cluster. A node can be added to more than one cluster.

        :return: Added node, for example:

            .. code-block:: json

                {
                    "cluster_oid": "ad6cfb53-3712-4683-8b15-f48a6d71d410",
                    "node_oid": "6009c4d1-b5e5-4ca8-aee3-0da28b5a08b2",
                    "hardlimit": null,
                    "softlimit": null
                }
        """
        self.log.info(
            '{func}(routercluster_oid={routercluster_oid}, node_oid={node_oid}, config={config}, details={details})',
            func=hltype(self.list_routercluster_nodes),
            routercluster_oid=hlid(routercluster_oid),
            node_oid=hlid(node_oid),
            config=config,
            details=details)

        config = config or {}
        config['cluster_oid'] = routercluster_oid
        config['node_oid'] = node_oid
        membership = RouterClusterNodeMembership.parse(config)

        with self.gdb.begin() as txn:
            node = self.gschema.nodes[txn, membership.node_oid]
            if not node or node.mrealm_oid != self._session._mrealm_oid:
                raise ApplicationError('crossbar.error.no_such_object',
                                       'no node with oid {} found'.format(membership.node_oid))

        with self.db.begin(write=True) as txn:
            routercluster = self.schema.routerclusters[txn, membership.cluster_oid]
            if not routercluster:
                raise ApplicationError('crossbar.error.no_such_object',
                                       'no routercluster with oid {} found'.format(membership.cluster_oid))

            self.schema.routercluster_node_memberships[txn, (membership.cluster_oid, membership.node_oid)] = membership

        res_obj = membership.marshal()
        self.log.info('node added to router cluster:\n{membership}', membership=res_obj)

        await self._session.publish('{}.on_routercluster_node_added'.format(self._prefix),
                                    res_obj,
                                    options=self._PUBOPTS)

        return res_obj

    @wamp.register(None, check_types=True)
    async def remove_routercluster_node(self,
                                        routercluster_oid: str,
                                        node_oid: str,
                                        details: Optional[CallDetails] = None) -> dict:
        """
        Remove a node from a router cluster.

        :param routercluster_oid: OID of the router cluster from which to remove the node.
        :param node_oid: OID of the node to remove from the router cluster

        :return: Node removed from router cluster, for example:

            .. code-block:: json

                {
                    "cluster_oid": "ad6cfb53-3712-4683-8b15-f48a6d71d410",
                    "node_oid": "6009c4d1-b5e5-4ca8-aee3-0da28b5a08b2",
                    "hardlimit": null,
                    "softlimit": null
                }
        """
        try:
            node_oid_ = uuid.UUID(node_oid)
        except Exception as e:
            raise ApplicationError('wamp.error.invalid_argument', 'invalid oid "{}"'.format(str(e)))

        try:
            routercluster_oid_ = uuid.UUID(routercluster_oid)
        except Exception as e:
            raise ApplicationError('wamp.error.invalid_argument', 'invalid oid "{}"'.format(str(e)))

        self.log.info('{func}(routercluster_oid={routercluster_oid}, node_oid={node_oid}, details={details})',
                      routercluster_oid=hlid(routercluster_oid_),
                      node_oid=hlid(node_oid),
                      func=hltype(self.remove_routercluster_node),
                      details=details)

        with self.gdb.begin() as txn:
            node = self.gschema.nodes[txn, node_oid_]
            if not node or node.mrealm_oid != self._session._mrealm_oid:
                raise ApplicationError('crossbar.error.no_such_object', 'no node with oid {} found'.format(node_oid_))

        with self.db.begin(write=True) as txn:
            routercluster = self.schema.routerclusters[txn, routercluster_oid_]
            if not routercluster:
                raise ApplicationError('crossbar.error.no_such_object',
                                       'no object with oid {} found'.format(routercluster_oid_))

            membership = self.schema.routercluster_node_memberships[txn, (routercluster_oid_, node_oid_)]
            if not membership:
                raise ApplicationError(
                    'crossbar.error.no_such_object',
                    'no association between node {} and routercluster {} found'.format(node_oid_, routercluster_oid_))

            del self.schema.routercluster_node_memberships[txn, (routercluster_oid_, node_oid_)]

        res_obj = membership.marshal()
        self.log.info('node removed from router cluster:\n{res_obj}', membership=res_obj)

        await self._session.publish('{}.on_routercluster_node_removed'.format(self._prefix),
                                    res_obj,
                                    options=self._PUBOPTS)

        return res_obj

    @wamp.register(None, check_types=True)
    def get_routercluster_node(self,
                               routercluster_oid: str,
                               node_oid: str,
                               details: Optional[CallDetails] = None) -> dict:
        """
        Get information (such as for example parallel degree) for the association
        of a node with a routercluster.

        :param routercluster_oid: The router cluster to which the node was added.
        :param node_oid: The node to return.

        :return: Information for the association of the node with the routercluster. For example:

            .. code-block:: json

                {
                    "cluster_oid": "ad6cfb53-3712-4683-8b15-f48a6d71d410",
                    "node_oid": "d87b502c-83d9-4cce-87be-0bb1bbd9539a",
                    "hardlimit": null,
                    "softlimit": null
                }
        """
        try:
            node_oid_ = uuid.UUID(node_oid)
        except Exception as e:
            raise ApplicationError('wamp.error.invalid_argument', 'invalid oid "{}"'.format(str(e)))

        try:
            routercluster_oid_ = uuid.UUID(routercluster_oid)
        except Exception as e:
            raise ApplicationError('wamp.error.invalid_argument', 'invalid oid "{}"'.format(str(e)))

        self.log.info('{func}(routercluster_oid={routercluster_oid}, node_oid={node_oid}, details={details})',
                      routercluster_oid=hlid(routercluster_oid_),
                      node_oid=hlid(node_oid_),
                      func=hltype(self.get_routercluster_node),
                      details=details)

        with self.gdb.begin() as txn:
            node = self.gschema.nodes[txn, node_oid_]
            if not node or node.mrealm_oid != self._session._mrealm_oid:
                raise ApplicationError('crossbar.error.no_such_object', 'no node with oid {} found'.format(node_oid_))

        with self.db.begin() as txn:
            routercluster = self.schema.routerclusters[txn, routercluster_oid_]
            if not routercluster:
                raise ApplicationError('crossbar.error.no_such_object',
                                       'no object with oid {} found'.format(routercluster_oid_))

            membership = self.schema.routercluster_node_memberships[txn, (routercluster_oid_, node_oid_)]
            if not membership:
                raise ApplicationError(
                    'crossbar.error.no_such_object',
                    'no association between node {} and routercluster {} found'.format(node_oid_, routercluster_oid_))

        res_obj = membership.marshal()

        return res_obj

    @wamp.register(None, check_types=True)
    def list_routercluster_workergroups(self,
                                        routercluster_oid: str,
                                        return_names: Optional[bool] = None,
                                        filter_by_status: Optional[str] = None,
                                        details: Optional[CallDetails] = None) -> List[str]:
        """
        List worker groups in a router cluster. Detail information for a router cluster worker group
        can be retrieved using :meth:`crossbar.master.cluster.routercluster.RouterClusterManager.get_routercluster_workergroup`.

        :param routercluster_oid: The object ID of the router cluster to list router worker groups for.
        :param return_names: If set, return router worker group names instead of object IDs.
        :param filter_by_status: If set, only return worker group in this status.

        :return: List of router cluster worker group object IDs, for example:

            .. code-block:: json

                [
                    "5c295684-7f7f-4560-b175-7466ed957c2e"
                ]

        or with ``return_names`` set:

            .. code-block:: json

                [
                    "mygroup1"
                ]
        """
        self.log.info('{func}(routercluster_oid={routercluster_oid}, details={details})',
                      routercluster_oid=hlid(routercluster_oid),
                      func=hltype(self.list_routercluster_workergroups),
                      details=details)

        try:
            routercluster_oid_ = uuid.UUID(routercluster_oid)
        except Exception as e:
            raise ApplicationError('wamp.error.invalid_argument', 'invalid oid "{}"'.format(str(e)))

        workergroups = []
        with self.db.begin() as txn:
            routercluster = self.schema.routerclusters[txn, routercluster_oid_]
            if not routercluster:
                raise ApplicationError('crossbar.error.no_such_object',
                                       'no object with oid {} found'.format(routercluster_oid))

            # Use index "idx_workergroup_by_cluster": (cluster_oid, workergroup_name) -> workergroup_oid
            from_key = (routercluster_oid_, '')
            to_key = (uuid.UUID(int=(int(routercluster_oid_) + 1)), '')
            for (_,
                 workergroup_name), workergroup_oid in self.schema.idx_workergroup_by_cluster.select(txn,
                                                                                                     from_key=from_key,
                                                                                                     to_key=to_key):
                if return_names:
                    workergroups.append(workergroup_name)
                else:
                    workergroups.append(str(workergroup_oid))

            return workergroups

    @wamp.register(None, check_types=True)
    async def add_routercluster_workergroup(self,
                                            routercluster_oid: str,
                                            workergroup: dict,
                                            details: Optional[CallDetails] = None) -> dict:
        """
        Add a Router worker group to a Router cluster. The ``workergroup`` can be configured:

        .. code-block:: json

            {
                "name": "mygroup1",
                "scale": 4
            }

        :param routercluster_oid: Router cluster to which to add the router worker group.
        :param workergroup: Web service definition object.

        :returns: Router cluster worker group creation information, for example:

            .. code-block:: json

                {
                    "changed": 1598452531613401997,
                    "cluster_oid": "b99833d5-0f03-4759-b1ed-b7059e81b2d8",
                    "description": null,
                    "label": null,
                    "name": "mygroup1",
                    "oid": "5c295684-7f7f-4560-b175-7466ed957c2e",
                    "scale": 4,
                    "status": "STOPPED",
                    "tags": null
                }
        """
        self.log.info('{func}(routercluster_oid={routercluster_oid}, workergroup={workergroup}, details={details})',
                      routercluster_oid=hlid(routercluster_oid),
                      workergroup=pformat(workergroup),
                      func=hltype(self.add_routercluster_workergroup),
                      details=details)

        try:
            routercluster_oid_ = uuid.UUID(routercluster_oid)
        except Exception as e:
            raise ApplicationError('wamp.error.invalid_argument', 'invalid oid "{}"'.format(str(e)))

        with self.db.begin(write=True) as txn:
            # get routercluster on which to create a new workergroup
            routercluster = self.schema.routerclusters[txn, routercluster_oid_]
            if not routercluster:
                raise ApplicationError('crossbar.error.no_such_object',
                                       'no object with oid {} found'.format(routercluster_oid_))

            # create new workergroup object
            workergroup_obj = RouterWorkerGroup.parse(workergroup)
            workergroup_obj.oid = uuid.uuid4()
            workergroup_obj.cluster_oid = routercluster_oid_
            workergroup_obj.status = WorkerGroupStatus.STOPPED
            workergroup_obj.changed = time_ns()

            # if no explicit workergroup name was given, auto-assign a name
            if not workergroup_obj.name:
                workergroup_obj.name = 'cwg_{}'.format(str(workergroup_obj.oid)[:8])

            # store workergroup in database
            self.schema.router_workergroups[txn, workergroup_obj.oid] = workergroup_obj
            self.log.info('New router worker group object stored in database:\n{workergroup}',
                          workergroup=pformat(workergroup_obj.marshal()))

            # collect all node OIDs currently associated with the router cluster
            # into a dict sorted by the current number of placements
            nodes = SortedDict()
            for _, node_oid in self.schema.routercluster_node_memberships.select(
                    txn,
                    from_key=(routercluster_oid_, uuid.UUID(bytes=b'\0' * 16)),
                    to_key=(uuid.UUID(int=(int(routercluster_oid_) + 1)), uuid.UUID(bytes=b'\0' * 16)),
                    return_values=False):

                # count current number of placements associated with given node
                cnt = 0
                for _ in self.schema.idx_workergroup_by_placement.select(
                        txn,
                        from_key=(workergroup_obj.cluster_oid, node_oid, uuid.UUID(bytes=b'\0' * 16)),
                        to_key=(workergroup_obj.cluster_oid, uuid.UUID(int=(int(node_oid) + 1)),
                                uuid.UUID(bytes=b'\0' * 16)),
                        return_keys=False):
                    cnt += 1

                # the dict is sorted ascending by cnt, that is nodes.peekitem(0)
                # will be a node with the smallest current number of placements
                nodes[node_oid] = cnt

            # create and store workergroup worker placements on nodes for the new workergroup
            for i in range(workergroup_obj.scale):
                # new placement on a node with smallest number of current placements
                placement_node_oid, placement_node_cnt = nodes.peekitem(0)

                self.log.info(
                    'Router worker placement selected node {placement_node_oid} with current worker count {placement_node_cnt}',
                    placement_node_oid=hlid(placement_node_oid),
                    placement_node_cnt=hlval(placement_node_cnt))

                placement = RouterWorkerGroupClusterPlacement()
                placement.oid = uuid.uuid4()
                placement.worker_group_oid = workergroup_obj.oid
                placement.cluster_oid = routercluster_oid_
                placement.node_oid = placement_node_oid
                placement.worker_name = '{}_{}'.format(workergroup_obj.name, i + 1)
                placement.status = WorkerGroupStatus.STOPPED
                placement.changed = time_ns()
                placement.tcp_listening_port = 0

                self.schema.router_workergroup_placements[txn, placement.oid] = placement

                # keep track of new placement in our sorted dict
                nodes[placement_node_oid] += 1

                self.log.info('New router worker group placement object stored in database:\n{placement}',
                              placement=pformat(placement.marshal()))

        res_obj = workergroup_obj.marshal()

        await self._session.publish('{}.on_workergroup_added'.format(self._prefix), res_obj, options=self._PUBOPTS)

        self.log.info('Management API event <on_workergroup_added> published:\n{res_obj}', res_obj=res_obj)

        return res_obj

    @wamp.register(None, check_types=True)
    async def remove_routercluster_workergroup(self,
                                               routercluster_oid: str,
                                               workergroup_oid: str,
                                               details: Optional[CallDetails] = None) -> dict:
        """
        Remove a router worker group from a router cluster.

        :param routercluster_oid: The object ID of the router cluster to remove a router worker group from.
        :param workergroup_oid: The object ID of the router worker group to remove.

        :return: Removed router worker group, for example:

            .. code-block:: json

                {
                    "changed": 1598455166431307344,
                    "cluster_oid": "b99833d5-0f03-4759-b1ed-b7059e81b2d8",
                    "description": null,
                    "label": null,
                    "name": "mygroup1",
                    "oid": "fa5498b7-c660-4a5b-81f5-95f9223a19f5",
                    "scale": 4,
                    "status": "STOPPED",
                    "tags": null
                }
        """
        self.log.info(
            '{func}(routercluster_oid={routercluster_oid}, workergroup_oid={workergroup_oid}, details={details})',
            routercluster_oid=hlid(routercluster_oid),
            workergroup_oid=hlid(workergroup_oid),
            func=hltype(self.remove_routercluster_workergroup),
            details=details)

        try:
            routercluster_oid_ = uuid.UUID(routercluster_oid)
        except Exception as e:
            raise ApplicationError('wamp.error.invalid_argument', 'invalid routercluster_oid "{}"'.format(str(e)))

        try:
            workergroup_oid_ = uuid.UUID(workergroup_oid)
        except Exception as e:
            raise ApplicationError('wamp.error.invalid_argument', 'invalid workergroup_oid "{}"'.format(str(e)))

        if details and details.caller_authid:
            with self.gdb.begin() as txn:
                caller_oid = self.gschema.idx_users_by_email[txn, details.caller_authid]
                if not caller_oid:
                    raise ApplicationError('wamp.error.no_such_principal',
                                           'no user found for authid "{}"'.format(details.caller_authid))
        else:
            raise ApplicationError('wamp.error.no_such_principal', 'cannot map user - no caller authid available')

        with self.db.begin(write=True) as txn:
            routercluster = self.schema.routerclusters[txn, routercluster_oid_]
            if routercluster:
                if routercluster.owner_oid != caller_oid:
                    raise ApplicationError('wamp.error.not_authorized',
                                           'only owner is allowed to modify router cluster')
                if routercluster.status != cluster.STATUS_STOPPED:
                    raise ApplicationError('crossbar.error.not_stopped')
            else:
                raise ApplicationError('crossbar.error.no_such_object',
                                       'no object with oid {} found'.format(routercluster_oid_))

            workergroup = self.schema.router_workergroups[txn, workergroup_oid_]
            if not workergroup:
                raise ApplicationError('crossbar.error.no_such_object',
                                       'no workergroup with oid {} found'.format(workergroup_oid_))

            # FIXME: check that the worker group has no running application realms
            # FIXME: remove all router worker group placements from self.schema.router_workergroup_placements

            del self.schema.router_workergroups[txn, workergroup_oid_]

        self.log.info('router cluster work group object removed from database:\n{workergroup}',
                      workergroup=workergroup)

        res_obj = workergroup.marshal()

        await self._session.publish('{}.on_workergroup_removed'.format(self._prefix), res_obj, options=self._PUBOPTS)

        self.log.info('Management API event <on_workergroup_removed> published:\n{res_obj}', res_obj=res_obj)

        return res_obj

    @wamp.register(None, check_types=True)
    def get_routercluster_workergroup(self,
                                      routercluster_oid: str,
                                      workergroup_oid: str,
                                      details: Optional[CallDetails] = None) -> dict:
        """
        Get definition of a router worker group in a cluster by ID.

        :param routercluster_oid: The router cluster running the router worker group to return.
        :param workergroup_oid: The router worker group to return.

        :return: The router cluster worker group, for example:

            .. code-block:: json

                {
                    "changed": 1598452531613401997,
                    "cluster_oid": "b99833d5-0f03-4759-b1ed-b7059e81b2d8",
                    "description": null,
                    "label": null,
                    "name": "mygroup1",
                    "oid": "5c295684-7f7f-4560-b175-7466ed957c2e",
                    "scale": 4,
                    "status": "STOPPED",
                    "tags": null
                }
        """
        self.log.info(
            '{func}(routercluster_oid={routercluster_oid}, workergroup_oid={workergroup_oid}, details={details})',
            routercluster_oid=hlid(routercluster_oid),
            workergroup_oid=hlid(workergroup_oid),
            func=hltype(self.get_routercluster_workergroup),
            details=details)

        try:
            routercluster_oid_ = uuid.UUID(routercluster_oid)
        except Exception as e:
            raise ApplicationError('wamp.error.invalid_argument', 'invalid oid "{}"'.format(str(e)))

        try:
            workergroup_oid_ = uuid.UUID(workergroup_oid)
        except Exception as e:
            raise ApplicationError('wamp.error.invalid_argument', 'invalid oid "{}"'.format(str(e)))

        with self.db.begin() as txn:
            workergroup = self.schema.router_workergroups[txn, workergroup_oid_]
            if not workergroup:
                raise ApplicationError('crossbar.error.no_such_object',
                                       'no worker group with oid {} found'.format(workergroup_oid_))

        if workergroup.cluster_oid != routercluster_oid_:
            raise ApplicationError(
                'crossbar.error.no_such_object',
                'worker group with oid {} found, but not associated with given router cluster'.format(
                    workergroup_oid_))

        return workergroup.marshal()

    @wamp.register(None, check_types=True)
    def get_routercluster_workergroup_by_name(self,
                                              routercluster_name: str,
                                              workergroup_name: str,
                                              details: Optional[CallDetails] = None) -> dict:
        """
        Get definition of a router worker group in a cluster by name.

        See also :meth:`crossbar.master.cluster.routercluster.RouterClusterManager.get_routercluster_workergroup`.

        :param routercluster_name: The router cluster running the router worker group to return.
        :param workergroup_name: The router worker group to return.

        :return: The router cluster worker group.
        """
        self.log.info(
            '{func}(routercluster_name="{routercluster_name}", workergroup_name="{workergroup_name}", details={details})',
            routercluster_name=hlval(routercluster_name),
            workergroup_name=hlval(workergroup_name),
            func=hltype(self.get_routercluster_workergroup),
            details=details)

        with self.db.begin() as txn:
            routercluster_oid = self.schema.idx_routerclusters_by_name[txn, routercluster_name]
            if not routercluster_oid:
                raise ApplicationError('crossbar.error.no_such_object',
                                       'no routercluster named {}'.format(routercluster_name))

            workergroup_oid = self.schema.idx_workergroup_by_cluster[txn, (routercluster_oid, workergroup_name)]
            if not workergroup_oid:
                raise ApplicationError('crossbar.error.no_such_object',
                                       'no workergroup named {}'.format(workergroup_name))

            workergroup = self.schema.router_workergroups[txn, workergroup_oid]
            assert workergroup

        return workergroup.marshal()

    @wamp.register(None, check_types=True)
    def stat_routercluster_workergroup(self,
                                       routercluster_oid: str,
                                       workergroup_oid: str,
                                       details: Optional[CallDetails] = None) -> dict:
        """
        *NOT YET IMPLEMENTED*

        Return current status and statistics for the router worker group.

        :param routercluster_oid: The router cluster running the web service to return status and statistics for.
        :param workergroup_oid: The worker group to return status and statistics for.

        :return: Current status and statistics information for the router worker group.
        """
        self.log.info(
            '{func}(routercluster_oid={routercluster_oid}, workergroup_oid={workergroup_oid}, details={details})',
            func=hltype(self.stat_routercluster_workergroup),
            routercluster_oid=hlid(routercluster_oid),
            workergroup_oid=hlid(workergroup_oid),
            details=details)

        raise NotImplementedError()



#####################################################################################
#
#  Copyright (c) Crossbar.io Technologies GmbH
#  SPDX-License-Identifier: EUPL-1.2
#
#####################################################################################

from twisted.trial.unittest import TestCase

from crossbar import _compat as compat


class NativeStringTestCase(TestCase):
    """
    Tests for C{crossbar._compat.native_string}.
    """
    def test_bytes_always_native(self):
        """
        C{native_string}, with a bytes input, will always give a str output.
        """
        self.assertEqual(type(compat.native_string(b"foo")), str)

    def test_unicode_not_allowed(self):
        """
        A unicode argument should never be allowed.
        """
        with self.assertRaises(ValueError):
            compat.native_string("bar")



###############################################################################
#
# Crossbar.io Master
# Copyright (c) Crossbar.io Technologies GmbH. Licensed under EUPLv1.2.
#
###############################################################################

from typing import Dict, List  # noqa

import re

from txaio import make_logger

from autobahn.wamp.types import RegisterOptions, SubscribeOptions, PublishOptions
from crossbar._util import hl, hlid, hluserid

__all__ = ('RemoteApi', )


class RemoteApi(object):

    PREFIX = 'unset'

    PROCS = {'node': [], 'worker': []}  # type: Dict[str, List]

    EVENTS = {'node': [], 'worker': []}  # type: Dict[str, List]

    log = make_logger()

    def register(self, session):
        """

        :param session: MrealmController
        :return:
        """
        regs = []

        # create forwards on node_id
        #
        def create_forward_by_node(local_uri, remote_uri):
            def forward(node_oid, *args, **kwargs):
                # remove the calling origin details (do not forward those 1:1 at least - FIXME)
                kwargs.pop('details', None)

                # map the node OID (given as UUID string) to the node WAMP authid for ..
                node_authid = session.map_node_oid_to_authid(node_oid)

                # .. creating the remote URI to be used on the management uplink
                _remote_uri = remote_uri.format(node_id=node_authid)

                self.log.debug(
                    'Forwarding CFC {forward_type} on mrealm {realm} to node "{node_authid}" [node_oid={node_oid}, local=<{local_uri}>, remote=<{remote_uri}>]',
                    forward_type=hl('CALL'),
                    local_uri=hlid(local_uri),
                    remote_uri=hlid(_remote_uri),
                    node_oid=hlid(node_oid),
                    node_authid=hluserid(node_authid),
                    realm=hluserid(session._realm))
                return session.call(_remote_uri, *args, **kwargs)

            return session.register(forward, local_uri, options=RegisterOptions(details_arg='details'))

        procs_by_node = self.PROCS.get('node', [])

        for proc in procs_by_node:
            if type(proc) == tuple:
                proc, rproc = proc
            else:
                rproc = proc
            local_uri = '{prefix}{proc}'.format(prefix=self.PREFIX, proc=proc)
            remote_uri = 'crossbarfabriccenter.node.{{node_id}}.{proc}'.format(proc=rproc)
            regs.append(create_forward_by_node(local_uri, remote_uri))

        # create forwards on node_id, worker_id
        #
        def create_forward_by_worker(local_uri, remote_uri):
            def forward(node_oid, worker_id, *args, **kwargs):
                # remove the calling origin details (do not forward those 1:1 at least - FIXME)
                kwargs.pop('details', None)

                # map the node OID (given as UUID string) to the node WAMP authid for ..
                node_authid = session.map_node_oid_to_authid(node_oid)

                # .. creating the remote URI to be used on the management uplink
                _remote_uri = remote_uri.format(node_id=node_authid, worker_id=worker_id)

                self.log.debug(
                    'Forwarding CFC {forward_type} mrealm {realm} to node "{node_authid}" and worker {worker_id} [node_oid={node_oid}, local=<{local_uri}>, remote=<{remote_uri}>]',
                    forward_type=hl('CALL'),
                    local_uri=hlid(local_uri),
                    remote_uri=hlid(_remote_uri),
                    node_oid=hlid(node_oid),
                    node_authid=hluserid(node_authid),
                    worker_id=hluserid(worker_id),
                    realm=hluserid(session._realm))
                return session.call(_remote_uri, *args, **kwargs)

            return session.register(forward, local_uri, options=RegisterOptions(details_arg='details'))

        procs_by_worker = self.PROCS.get('worker', [])

        for proc in procs_by_worker:
            if type(proc) == tuple:
                proc, rproc = proc
            else:
                rproc = proc
            local_uri = '{prefix}{proc}'.format(prefix=self.PREFIX, proc=proc)
            remote_uri = 'crossbarfabriccenter.node.{{node_id}}.worker.{{worker_id}}.{proc}'.format(proc=rproc)
            regs.append(create_forward_by_worker(local_uri, remote_uri))

        return regs

    def subscribe(self, session):
        subs = []

        #
        # create forwards on node_id
        #
        def create_forward_by_node(local_uri, remote_uri, remote_uri_regex):
            pat = re.compile(remote_uri_regex)

            def forward(*args, **kwargs):
                details = kwargs.pop('details', None)
                if details:
                    match = pat.match(details.topic)
                    if match:
                        node_id = match.groups()[0]

                        # FIXME: map back from node authid (?) to node OID (as UUID string)?

                        self.log.debug(
                            'Forwarding CFC {forward_type} on mrealm {realm} from node {node_id} [local=<{local_uri}>, remote=<{remote_uri}>]',
                            forward_type=hl('EVENT'),
                            local_uri=hlid(local_uri),
                            remote_uri=hlid(details.topic),
                            node_id=hluserid(node_id),
                            realm=hluserid(session._realm))

                        return session.publish(local_uri,
                                               node_id,
                                               *args,
                                               **kwargs,
                                               options=PublishOptions(exclude_me=False))

                # should not arrive here
                session.log.warn(
                    'received unexpected event to forward for management API: local_uri={local_uri}, remote_uri={remote_uri}, remote_uri_regex={remote_uri_regex} details={details}',
                    local_uri=local_uri,
                    remote_uri=remote_uri,
                    remote_uri_regex=remote_uri_regex,
                    details=details)

            return session.subscribe(forward, remote_uri, SubscribeOptions(match='wildcard', details=True))

        topics_by_node = self.EVENTS.get('node', [])

        for topic in topics_by_node:
            if type(topic) == tuple:
                topic, rtopic = topic
            else:
                rtopic = topic
            local_uri = '{prefix}{topic}'.format(prefix=self.PREFIX, topic=topic)
            remote_uri = 'crossbarfabriccenter.node..{topic}'.format(topic=rtopic)
            remote_uri_regex = r'^crossbarfabriccenter.node.([a-z0-9][a-z0-9_\-]*).{topic}$'.format(topic=rtopic)
            subs.append(create_forward_by_node(local_uri, remote_uri, remote_uri_regex))

        #
        # create forwards on node_id, worker_id
        #
        def create_forward_by_worker(local_uri, remote_uri, remote_uri_regex):
            pat = re.compile(remote_uri_regex)

            def forward(*args, **kwargs):
                details = kwargs.pop('details', None)
                if details:
                    match = pat.match(details.topic)
                    if match:
                        node_id, worker_id = match.groups()

                        # FIXME: map back from node authid (?) to node OID (as UUID string)?

                        self.log.debug(
                            'Forwarding CFC {forward_type} on mrealm {realm} from node {node_id} and worker {worker_id} [local=<{local_uri}>, remote=<{remote_uri}>]',
                            forward_type=hl('EVENT'),
                            local_uri=hlid(local_uri),
                            remote_uri=hlid(details.topic),
                            node_id=hluserid(node_id),
                            worker_id=hluserid(worker_id),
                            realm=hluserid(session._realm))

                        return session.publish(local_uri,
                                               node_id,
                                               worker_id,
                                               *args,
                                               **kwargs,
                                               options=PublishOptions(exclude_me=False))

                # should not arrive here
                session.log.warn(
                    'received unexpected event to forward for management API: local_uri={local_uri}, remote_uri={remote_uri}, remote_uri_regex={remote_uri_regex} details={details}',
                    local_uri=local_uri,
                    remote_uri=remote_uri,
                    remote_uri_regex=remote_uri_regex,
                    details=details)

            return session.subscribe(forward, remote_uri, SubscribeOptions(match='wildcard', details=True))

        topics_by_worker = self.EVENTS.get('worker', [])

        for topic in topics_by_worker:
            if type(topic) == tuple:
                topic, rtopic = topic
            else:
                rtopic = topic
            local_uri = '{prefix}{topic}'.format(prefix=self.PREFIX, topic=topic)
            remote_uri = 'crossbarfabriccenter.node..worker..{topic}'.format(topic=rtopic)
            remote_uri_regex = r'^crossbarfabriccenter.node.([a-z0-9][a-z0-9_\-]*).worker.([a-z0-9][a-z0-9_\-]*).{topic}$'.format(
                topic=rtopic)
            subs.append(create_forward_by_worker(local_uri, remote_uri, remote_uri_regex))

        return subs



#####################################################################################
#
#  Copyright (c) Crossbar.io Technologies GmbH
#  SPDX-License-Identifier: EUPL-1.2
#
#####################################################################################

from txaio import make_logger

from twisted.web import server
from twisted.web.http import HTTPChannel


def createHSTSRequestFactory(requestFactory, hstsMaxAge=31536000):
    """
    Builds a request factory that sets HSTS (HTTP Strict Transport
    Security) headers, by wrapping another request factory.
    """
    def makeRequest(*a, **kw):
        request = requestFactory(*a, **kw)
        request.responseHeaders.setRawHeaders("Strict-Transport-Security", ["max-age={}".format(hstsMaxAge)])
        return request

    return makeRequest


class _LessNoisyHTTPChannel(HTTPChannel):
    """
    Internal helper.

    This is basically exactly what Twisted does, except without using
    "log.msg" so we can put it at debug log-level instead
    """
    log = make_logger()

    def timeoutConnection(self):
        self.log.debug(
            "Timing out client: {peer}",
            peer=self.transport.getPeer(),
        )
        if self.abortTimeout is not None:
            self._abortingCall = self.callLater(self.abortTimeout, self.forceAbortClient)
        self.loseConnection()


class Site(server.Site):
    def __init__(self,
                 resource,
                 client_timeout=None,
                 access_log=None,
                 display_tracebacks=None,
                 hsts=None,
                 hsts_max_age=None):

        server.Site.__init__(self, resource, timeout=client_timeout)

        # Web access logging
        if not access_log:
            self.noisy = False
            self.log = lambda _: None

        # Traceback rendering
        self.displayTracebacks = True if display_tracebacks else False

        # HSTS
        if hsts:
            hsts_max_age = hsts_max_age or 31536000
            self.requestFactory = createHSTSRequestFactory(self.requestFactory, hsts_max_age)



#####################################################################################
#
#  Copyright (c) Crossbar.io Technologies GmbH
#  SPDX-License-Identifier: EUPL-1.2
#
#####################################################################################

# IMPORTANT:
#
# 1. You must use pip 22.0.3 or later
#       pls https://github.com/crossbario/crossbar/pull/1943#issuecomment-1037569885
#
# 2. No, we can't add that pip version requirement here
#       see https://stackoverflow.com/a/60159094/884770
#

import os

from setuptools import setup, find_packages


# read package description
with open('README.rst') as f:
    long_description = f.read()

# read package version
with open('crossbar/_version.py') as f:
    exec(f.read())  # defines __version__

install_requires = []

# https://mike.zwobble.org/2013/05/adding-git-or-hg-or-svn-dependencies-in-setup-py/
dependency_links = []

if False:
    extras_require = {}
else:
    extras_require = {
        'dev': []
    }
    with open('requirements-dev.txt') as f:
        for line in f.read().splitlines():
            extras_require['dev'].append(line.strip())

with open('requirements-min.txt') as f:
    for line in f.read().splitlines():
        line = line.strip()
        if not line.startswith('#'):
            parts = line.strip().split(';')
            if len(parts) > 1:
                parts[0] = parts[0].strip()
                parts[1] = ':{}'.format(parts[1].strip())
                if parts[1] not in extras_require:
                    extras_require[parts[1]] = []
                extras_require[parts[1]].append(parts[0])
            else:
                name = parts[0]
                # do NOT (!) touch this!
                # https://mike.zwobble.org/2013/05/adding-git-or-hg-or-svn-dependencies-in-setup-py/
                if name.startswith('git+'):
                    dependency_links.append(name)
                else:
                    install_requires.append(name)

# enforce use of CFFI for LMDB
os.environ['LMDB_FORCE_CFFI'] = '1'

# enforce use of bundled libsodium
os.environ['SODIUM_INSTALL'] = 'bundled'

# enforce use of pure Python py-ubjson (no Cython)
os.environ['PYUBJSON_NO_EXTENSION'] = '1'

# now actually call into setuptools ..
setup(
    name='crossbar',
    version=__version__,
    description='Crossbar.io multi-protocol (WAMP/WebSocket, REST/HTTP, MQTT) application router for microservices.',
    long_description=long_description,
    author='Crossbar.io Technologies GmbH',
    url='http://crossbar.io/',
    platforms=('Any'),
    license="European Union Public Licence 1.2 (EUPL 1.2)",
    install_requires=install_requires,

    # https://mike.zwobble.org/2013/05/adding-git-or-hg-or-svn-dependencies-in-setup-py/
    dependency_links=dependency_links,

    extras_require=extras_require,
    entry_points={
        # CLI entry function
        'console_scripts': [
            'crossbar = crossbar:run'
        ]
    },
    packages=find_packages(),
    include_package_data=True,
    data_files=[('.', ['crossbar/LEGAL', 'crossbar/LICENSE', 'crossbar/LICENSE-FOR-API', 'crossbar/LICENSES-OSS', 'crossbar.ico'])],
    zip_safe=False,
    python_requires='>=3.7',

    # http://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=["License :: OSI Approved :: European Union Public Licence 1.2 (EUPL 1.2)",
                 "Development Status :: 5 - Production/Stable",
                 "Environment :: No Input/Output (Daemon)",
                 "Environment :: Console",
                 "Framework :: Twisted",
                 "Intended Audience :: Developers",
                 "Operating System :: OS Independent",
                 "Programming Language :: Python :: 3.7",
                 "Programming Language :: Python :: 3.8",
                 "Programming Language :: Python :: 3.9",
                 "Programming Language :: Python :: 3.10",
                 "Programming Language :: Python :: Implementation :: CPython",
                 "Programming Language :: Python :: Implementation :: PyPy",
                 "Topic :: Internet",
                 "Topic :: Internet :: WWW/HTTP :: HTTP Servers",
                 "Topic :: Communications",
                 "Topic :: Database",
                 "Topic :: Home Automation",
                 "Topic :: Software Development :: Libraries",
                 "Topic :: Software Development :: Libraries :: Application Frameworks",
                 "Topic :: Software Development :: Embedded Systems",
                 "Topic :: Software Development :: Object Brokering",
                 "Topic :: System :: Distributed Computing",
                 "Topic :: System :: Networking"],
    keywords='crossbar router autobahn autobahn.ws websocket realtime rfc6455 wamp rpc pubsub oracle postgres postgresql'
)



#####################################################################################
#
#  Copyright (c) Crossbar.io Technologies GmbH
#  SPDX-License-Identifier: EUPL-1.2
#
#####################################################################################

import contextlib
import socket
import sys
import json
import os
import re
import inspect
import uuid
import copy
from collections.abc import Mapping

import click

from autobahn.wamp import CallDetails
from crossbar.common.checkconfig import InvalidConfigException

_ENVPAT_STR = r'^\$\{(.+)\}$'
_ENVPAT = re.compile(_ENVPAT_STR)

DEBUG_LIFECYCLE = False
DEBUG_PROGRAMFLOW = False


def set_flags_from_args(_args):
    global DEBUG_LIFECYCLE
    global DEBUG_PROGRAMFLOW

    for arg in _args:
        if arg.strip().lower() == '--debug-lifecycle':
            DEBUG_LIFECYCLE = True
        if arg.strip().lower() == '--debug-programflow':
            DEBUG_PROGRAMFLOW = True


# FS path to controlling terminal
_TERMINAL = None

# Linux, *BSD and MacOSX
if sys.platform.startswith('linux') or 'bsd' in sys.platform or sys.platform.startswith('darwin'):
    _TERMINAL = '/dev/tty' if os.path.exists('/dev/tty') else None
# Windows
elif sys.platform in ['win32']:
    pass
# Other OS
else:
    pass

# still, we might not be able to use TTY, so duck test it:
if _TERMINAL:
    try:
        with open('/dev/tty', 'w') as f:
            f.write('\n')
            f.flush()
    except:
        # under systemd: OSError: [Errno 6] No such device or address: '/dev/tty'
        _TERMINAL = None


def class_name(obj):
    """
    This returns a name like "module.Class" given either an instance, or a class.
    """

    if inspect.isclass(obj):
        cls = obj
    else:
        cls = obj.__class__
    return '{}.{}'.format(cls.__module__, cls.__name__)


def dump_json(obj, minified=True):
    """
    Dump JSON to a string, either pretty printed or not. Returns a Unicode
    string.
    """
    if minified:
        return json.dumps(obj, separators=(',', ':'), ensure_ascii=False)

    else:
        return json.dumps(obj, indent=4, separators=(',', ': '), sort_keys=False, ensure_ascii=False)


def hl(text, bold=False, color='yellow'):
    """
    Returns highlighted text.
    """
    if not isinstance(text, str):
        text = '{}'.format(text)
    return click.style(text, fg=color, bold=bold)


def _qn(obj):
    if inspect.isclass(obj) or inspect.isfunction(obj) or inspect.ismethod(obj):
        qn = '{}.{}'.format(obj.__module__, obj.__qualname__)
    else:
        qn = 'unknown'
    return qn


# def hltype(obj, render=DEBUG_PROGRAMFLOW):
def hltype(obj, render=True):

    if render:
        qn = _qn(obj).split('.')
        text = hl(qn[0], color='yellow', bold=True) + hl('.' + '.'.join(qn[1:]), color='yellow', bold=False)
        return '<' + text + '>'
    else:
        return ''


def hlflag(flag, true_txt='YES', false_txt='NO', null_txt='UNSET'):
    assert flag is None or type(flag) == bool
    if flag is None:
        return hl('{}'.format(null_txt), color='blue', bold=True)
    elif flag:
        return hl('{}'.format(true_txt), color='green', bold=True)
    else:
        return hl('{}'.format(false_txt), color='red', bold=True)


def hlid(oid):
    return hl('{}'.format(oid), color='blue', bold=True)


def hlval(val, color='white'):
    return hl('{}'.format(val), color=color, bold=True)


def hluserid(oid):
    """
    Returns highlighted text.
    """
    if not isinstance(oid, str):
        oid = '{}'.format(oid)
    return hl('"{}"'.format(oid), color='yellow', bold=True)


def hlfixme(msg, obj):
    return hl('FIXME: {} {}'.format(msg, _qn(obj)), color='green', bold=True)


def hlcontract(oid):
    if not isinstance(oid, str):
        oid = '{}'.format(oid)
    return hl('<{}>'.format(oid), color='magenta', bold=True)


def term_print(text):
    """
    This directly prints to the process controlling terminal (if there is any).
    It bypasses any output redirections, or closes stdout/stderr pipes.

    This can be used eg for "admin messages", such as "node is shutting down now!"

    This currently only works on Unix like systems (tested only on Linux).
    When it cannot do so, it falls back to plain old print.
    """
    if DEBUG_LIFECYCLE:
        text = '{:<44}'.format(text)
        text = click.style(text, fg='blue', bold=True)
        if _TERMINAL:
            with open('/dev/tty', 'w') as f:
                f.write(text + '\n')
                f.flush()
        else:
            print(text)


def _add_debug_options(parser):
    parser.add_argument('--debug-lifecycle',
                        action='store_true',
                        help="This debug flag enables overall program lifecycle messages directly to terminal.")

    parser.add_argument(
        '--debug-programflow',
        action='store_true',
        help="This debug flag enables program flow log messages with fully qualified class/method names.")

    return parser


def _add_cbdir_config(parser):
    parser.add_argument('--cbdir',
                        type=str,
                        default=None,
                        help="Crossbar.io node directory (overrides ${CROSSBAR_DIR} and the default ./.crossbar)")

    parser.add_argument('--config',
                        type=str,
                        default=None,
                        help="Crossbar.io configuration file (overrides default CBDIR/config.json)")

    return parser


def _add_log_arguments(parser):
    color_args = dict({
        "type": str,
        "default": "auto",
        "choices": ["true", "false", "auto"],
        "help": "If logging should be colored."
    })
    parser.add_argument('--color', **color_args)

    log_level_args = dict({
        "type": str,
        "default": 'info',
        "choices": ['none', 'error', 'warn', 'info', 'debug', 'trace'],
        "help": ("How much Crossbar.io should log to the terminal, in order of verbosity.")
    })
    parser.add_argument('--loglevel', **log_level_args)

    parser.add_argument('--logformat',
                        type=str,
                        default='standard',
                        choices=['syslogd', 'standard', 'none'],
                        help=("The format of the logs -- suitable for syslogd, not colored, or colored."))

    parser.add_argument('--logdir',
                        type=str,
                        default=None,
                        help="Crossbar.io log directory (default: <Crossbar Node Directory>/)")

    parser.add_argument('--logtofile', action='store_true', help="Whether or not to log to file")

    return parser


def get_free_tcp_port(host='127.0.0.1'):
    """
    Returns random, free listening port.

    .. note::

        This is _not_ completely race free, as a port returned as free is closed
        before returning, and might then be used by another process before the caller
        of this function can actually bind it. So watch out ..

    :param host: Host (interface) for which to return a free port for.
    :type host: str

    :return: Free TCP listening port.
    :rtype: int
    """
    with contextlib.closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as sock:
        sock.bind((host, 0))
        return sock.getsockname()[1]


def first_free_tcp_port(host='127.0.0.1', portrange=(1024, 65535)):
    """
    Returns the first free listening port within the given range.

    :param host: Host (interface) for which to return a free port for.
    :type host: str

    :param portrange: Pair of start and end port for port range to select free port within.
    :type portrange: tuple

    :return: Free TCP listening port.
    :rtype: int
    """
    port, max_port = portrange
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    while port <= max_port:
        try:
            sock.bind((host, port))
            sock.close()
            return port
        except OSError:
            port += 1
    raise IOError('no free ports')


def get_free_tcp_address(host='127.0.0.1'):
    """
    Returns default local listening address with random port.

    Note: this is _not_ completely race free, as a port returned as free
    might be used by another process before the caller can bind it.

    :return: Default/free listening address:port.
    :rtype: str
    """
    # source: https://gist.github.com/gabrielfalcao/20e567e188f588b65ba2

    tcp = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    tcp.bind((host, 0))
    host, port = tcp.getsockname()
    tcp.close()
    address = 'tcp://{host}:{port}'.format(**locals())

    return address


def _deep_merge_map(a, b):
    """

    :param a:
    :param b:
    :return:
    """
    assert isinstance(a, Mapping)
    assert isinstance(b, Mapping)

    new_map = copy.deepcopy(a)

    for k, v in b.items():
        if v is None:
            if k in new_map and new_map[k]:
                del new_map[k]
        else:
            if k in new_map and isinstance(new_map[k], Mapping):
                assert isinstance(v, Mapping)
                new_map[k] = _deep_merge_map(new_map[k], v)

            # use list, not Sequence, since strings are also Sequences!
            # elif k in new_map and isinstance(new_map[k], Sequence):
            elif k in new_map and isinstance(new_map[k], list):
                # assert isinstance(v, Sequence)
                assert isinstance(v, list)

                new_map[k] = _deep_merge_list(new_map[k], v)

            else:
                new_map[k] = v

    return new_map


def _deep_merge_list(a, b):
    """
    Merges two lists. The list being merged must have length >= the
    list into which is merged.


    :param a: The list into which the other list is merged
    :param b: The list to be merged into the former
    :return: The merged list
    """
    # assert isinstance(a, Sequence)
    # assert isinstance(b, Sequence)
    assert isinstance(a, list)
    assert isinstance(b, list)

    assert len(b) >= len(a)
    if len(b) > len(a):
        for i in range(len(a), len(b)):
            assert b[i] is not None
            assert b[i] != 'COPY'

    new_list = []
    i = 0
    for item in b:
        if item is None:
            # drop the item from the target list
            pass
        elif item == 'COPY':
            # copy the item to the target list
            new_list.append(a[i])
        else:
            if i < len(a):
                # add merged item
                new_list.append(_deep_merge_object(a[i], item))
            else:
                # add new item
                new_list.append(item)
        i += 1
    return new_list


def _deep_merge_object(a, b):
    """

    :param a:
    :param b:
    :return:
    """
    # if isinstance(a, Sequence):
    if isinstance(a, list):
        return _deep_merge_list(a, b)
    elif isinstance(a, Mapping):
        return _deep_merge_map(a, b)
    else:
        return b


def merge_config(base_config, other_config):
    """

    :param base_config:
    :param other_config:
    :return:
    """
    if not isinstance(base_config, Mapping):
        raise InvalidConfigException('invalid type for configuration item - expected dict, got {}'.format(
            type(base_config).__name__))

    if not isinstance(other_config, Mapping):
        raise InvalidConfigException('invalid type for configuration item - expected dict, got {}'.format(
            type(other_config).__name__))

    merged_config = copy.deepcopy(base_config)

    if 'controller' in other_config:
        merged_config['controller'] = _deep_merge_map(merged_config.get('controller', {}), other_config['controller'])

    if 'workers' in other_config:
        merged_config['workers'] = _deep_merge_list(base_config.get('workers', []), other_config['workers'])

    return merged_config


def extract_member_oid(details: CallDetails) -> uuid.UUID:
    """
    Extract the XBR network member ID from the WAMP session authid (eg ``member_oid==72de3e0c-ca62-452f-8f09-2d3d30d1b511`` from ``authid=="member-72de3e0c-ca62-452f-8f09-2d3d30d1b511"``

    :param details: Call details.

    :return: Extracted XBR network member ID.
    """
    if details and details.caller_authrole == 'member' and details.caller_authid:
        return uuid.UUID(details.caller_authid[7:])
    else:
        raise RuntimeError('no XBR member identification in call details\n{}'.format(details))


def alternative_username(username):
    max_i = None
    for i in range(len(username) - 1, -1, -1):
        if username[i:].isdigit():
            max_i = i
    if max_i is not None:
        next = int(username[max_i:]) + 1
        alt_username = '{}{}'.format(username[:max_i], next)
    else:
        alt_username = '{}{}'.format(username, 1)
    return alt_username


def maybe_from_env(value):
    if isinstance(value, str):
        match = _ENVPAT.match(value)
        if match and match.groups():
            var = match.groups()[0]
            if var in os.environ:
                new_value = os.environ[var]
                return True, new_value
            else:
                print(
                    'WARNING: environment variable "{}" not set, but needed in XBR backend configuration'.format(var))
                return False, None
        else:
            return False, value
    else:
        return False, value



##############################################################################
#
#                        Crossbar.io
#     Copyright (C) Crossbar.io Technologies GmbH. All rights reserved.
#
##############################################################################

import os
from copy import deepcopy
from pprint import pformat

from twisted.internet.defer import inlineCallbacks
from twisted.web import server
from twisted.web.resource import Resource

import crossbar
from crossbar._util import hl
from crossbar.webservice.base import RouterWebService

from txaio import time_ns, make_logger

__all__ = ('RouterWebServiceRegisterMe', )


class RegisterMeResource(Resource):
    """
    Web resource to register master nodes with the XBR network.

    IMPORTANT: this resource is only indented to run inside master nodes!
    """
    log = make_logger()

    register_at = 'https://planet.xbr.network'

    def __init__(self, templates, worker, node_type=None):
        Resource.__init__(self)
        self._worker = worker
        self._service_realm = 'com.crossbario.fabric'
        self._service_agent = self._worker.realm_by_name(self._service_realm).session

        self._page = templates.get_template('registerme.html')
        self._pid = '{}'.format(os.getpid())
        self._node_type = node_type.strip().upper()

        assert self._node_type in ['MASTER']

    def _delayedRender(self, infos, request):
        try:
            peer = request.transport.getPeer()
            peer = '{}:{}'.format(peer.host, peer.port)
        except:
            peer = '?:?'

        kwargs = deepcopy(infos)

        node_time = time_ns()
        register_url = '{}/register-node?node_type={}&node_key={}&node_time={}'.format(
            self.register_at, self._node_type, kwargs['node_status']['pubkey'], node_time)

        kwargs['node_time'] = node_time
        kwargs['is_registered'] = False
        kwargs['cb_version'] = crossbar.__version__
        kwargs['worker_pid'] = self._pid
        kwargs['peer'] = peer
        kwargs['register_url'] = register_url
        kwargs['node_type'] = self._node_type

        s = self._page.render(**kwargs)
        request.write(s.encode('utf8'))
        request.finish()

    @inlineCallbacks
    def _do_get(self, request):
        try:
            node_status = yield self._worker.call('crossbar.get_status')
            master_status = yield self._service_agent.call('crossbarfabriccenter.domain.get_status')
            master_license = yield self._service_agent.call('crossbarfabriccenter.domain.get_license')
            master_version = yield self._service_agent.call('crossbarfabriccenter.domain.get_version')
            infos = {
                'node_status': node_status,
                'master_status': master_status,
                'master_license': master_license,
                'master_version': master_version,
            }
            self.log.debug('rendering {page} page using data\n{data}',
                           page=hl('register-master', bold=True),
                           data=pformat(infos))
            self._delayedRender(infos, request)
        except:
            self.log.failure()
            request.finish()

    def render_GET(self, request):
        self._do_get(request)
        return server.NOT_DONE_YET

    def getChild(self, path, request):
        self.log.debug('{kass}.getChild(path={path}, request={request}, prepath={prepath}, postpath={postpath})',
                       kass=self.__class__.__name__,
                       path=path,
                       prepath=request.prepath,
                       postpath=request.postpath,
                       request=request)

        search_path = b'/'.join([path] + request.postpath).decode('utf8')

        if search_path == '' or search_path.endswith('/') or search_path in ['register', 'register.html']:
            return self
        else:
            return Resource.getChild(self, path, request)


class RouterWebServiceRegisterMe(RouterWebService):
    """
    Crossbar.io "register-me" home page.

    IMPORTANT: this web service is only indented to run inside master nodes!
    """
    @staticmethod
    def check(personality, config):
        """
        Checks the configuration item. When errors are found, an
        InvalidConfigException exception is raised.

        :param personality: The node personality class.
        :param config: The Web service configuration item.
        :raises: crossbar.common.checkconfig.InvalidConfigException
        """
        pass

    @staticmethod
    def create(transport, path, config):
        personality = transport.worker.personality
        personality.WEB_SERVICE_CHECKERS['registerme'](personality, config)

        resource = RegisterMeResource(transport.templates, transport.worker, node_type=personality.NAME)

        return RouterWebServiceRegisterMe(transport, path, config, resource)



#####################################################################################
#
#  Copyright (c) Crossbar.io Technologies GmbH
#  SPDX-License-Identifier: EUPL-1.2
#
#####################################################################################

from crossbar.bridge.rest.callee import RESTCallee
from crossbar.bridge.rest.caller import CallerResource
from crossbar.bridge.rest.publisher import PublisherResource
from crossbar.bridge.rest.subscriber import MessageForwarder
from crossbar.bridge.rest.webhook import WebhookResource

__all__ = ("PublisherResource", "CallerResource", "RESTCallee", "MessageForwarder", "WebhookResource")



##############################################################################
#
#                        Crossbar.io
#     Copyright (C) Crossbar.io Technologies GmbH. All rights reserved.
#
##############################################################################

from typing import Dict
from collections.abc import Mapping, Sequence
from pprint import pformat

import six
import txaio

from crossbar.personality import Personality as CrossbarPersonality
from crossbar.personality import default_native_workers
from crossbar.common import checkconfig
from crossbar.node.node import NodeOptions
from crossbar.node.worker import RouterWorkerProcess

from crossbar.edge.node.node import FabricNode
from crossbar.edge.worker.realmstore import RealmStoreDatabase
from crossbar.edge.worker.router import ExtRouterController
from crossbar.edge.worker.hostmonitor import HostMonitor, HostMonitorProcess
from crossbar.edge.worker.xbrmm import MarketplaceController, MarketplaceControllerProcess
from crossbar.edge.webservice import RouterWebServicePairMe


def do_nothing(*args, **kw):
    return


# check blockchain configuration item (can be part of controller and market maker configurations)
def check_blockchain(personality, blockchain):
    # Examples:
    #
    # "blockchain": {
    #     "type": "ethereum",
    #     "gateway": {
    #         "type": "auto"
    #     }
    # }
    #
    # "blockchain": {
    #     "type": "ethereum",
    #     "gateway": {
    #         "type": "user",
    #         "http": "http://127.0.0.1:8545"
    #         "websocket": "ws://127.0.0.1:8545"
    #     },
    #     "from_block": 1,
    #     "chain_id": 5777
    # }
    #
    # "blockchain": {
    #     "type": "ethereum",
    #     "gateway": {
    #         "type": "infura",
    #         "network": "ropsten",
    #         "key": "${INFURA_PROJECT_ID}",
    #         "secret": "${INFURA_PROJECT_SECRET}"
    #     },
    #     "from_block": 6350652,
    #     "chain_id": 3
    # }
    checkconfig.check_dict_args(
        {
            'id': (False, [six.text_type]),
            'type': (True, [six.text_type]),
            'gateway': (True, [Mapping]),
            'key': (False, [six.text_type]),
            'from_block': (False, [int]),
            'chain_id': (False, [int]),
        }, blockchain, "blockchain configuration item {}".format(pformat(blockchain)))

    if blockchain['type'] not in ['ethereum']:
        raise checkconfig.InvalidConfigException('invalid type "{}" in blockchain configuration'.format(
            blockchain['type']))

    gateway = blockchain['gateway']
    if 'type' not in gateway:
        raise checkconfig.InvalidConfigException(
            'missing type in gateway item "{}" of blockchain configuration'.format(pformat(gateway)))

    if gateway['type'] not in ['infura', 'user', 'auto']:
        raise checkconfig.InvalidConfigException(
            'invalid type "{}" in gateway item of blockchain configuration'.format(gateway['type']))

    if gateway['type'] == 'infura':
        checkconfig.check_dict_args(
            {
                'type': (True, [six.text_type]),
                'network': (True, [six.text_type]),
                'key': (True, [six.text_type]),
                'secret': (True, [six.text_type]),
            }, gateway, "blockchain gateway configuration {}".format(pformat(gateway)))

        # allow to set value from environment variable
        gateway['key'] = checkconfig.maybe_from_env('blockchain.gateway["infura"].key',
                                                    gateway['key'],
                                                    hide_value=True)
        gateway['secret'] = checkconfig.maybe_from_env('blockchain.gateway["infura"].secret',
                                                       gateway['secret'],
                                                       hide_value=True)

    elif gateway['type'] == 'user':
        checkconfig.check_dict_args(
            {
                'type': (True, [six.text_type]),
                'http': (True, [six.text_type]),
                # 'websocket': (True, [six.text_type]),
            },
            gateway,
            "blockchain gateway configuration {}".format(pformat(gateway)))

    elif gateway['type'] == 'auto':
        checkconfig.check_dict_args({
            'type': (True, [six.text_type]),
        }, gateway, "blockchain gateway configuration {}".format(pformat(gateway)))

    else:
        # should not arrive here
        raise Exception('logic error')


# check database configuration item (can be part of controller, markets worker and market maker configurations)
def check_database(personality, database):
    checkconfig.check_dict_args(
        {
            'type': (True, [six.text_type]),
            'path': (True, [six.text_type]),
            'maxsize': (False, six.integer_types),
        }, database, "database configuration")

    if database['type'] not in ['cfxdb']:
        raise checkconfig.InvalidConfigException('invalid type "{}" in database configuration'.format(
            database['type']))

    if 'maxsize' in database:
        # maxsize must be between 1MB and 1TB
        if database['maxsize'] < 2**20 or database['maxsize'] > 2**40:
            raise checkconfig.InvalidConfigException(
                'invalid maxsize {} in database configuration - must be between 1MB and 1TB'.format(
                    database['maxsize']))


def check_controller_fabric(personality, fabric):
    """
    Check controller Fabric configuration override (which essentially is only
    for debugging purposes or for people running Crossbar.io Service on-premise)

    :param fabric: The Fabric configuration to check.
    :type fabric: dict
    """
    if not isinstance(fabric, Mapping):
        raise checkconfig.InvalidConfigException(
            "'fabric' in controller configuration must be a dictionary ({} encountered)\n\n".format(type(fabric)))

    for k in fabric:
        if k not in ['transport', 'heartbeat']:
            raise checkconfig.InvalidConfigException(
                "encountered unknown attribute '{}' in 'fabric' in controller configuration".format(k))

    if 'transport' in fabric:
        checkconfig.check_connecting_transport(personality, fabric['transport'])

    if 'heartbeat' in fabric:
        heartbeat = fabric['heartbeat']
        checkconfig.check_dict_args(
            {
                'startup_delay': (False, [int, float]),
                'heartbeat_period': (False, [int, float]),
                'include_system_stats': (False, [bool]),
                'send_workers_heartbeats': (False, [bool]),
                'aggregate_workers_heartbeats': (False, [bool]),
            }, heartbeat, "heartbeat configuration: {}".format(pformat(heartbeat)))


def check_controller(personality, controller, ignore=[]):
    res = checkconfig.check_controller(personality, controller, ['fabric', 'blockchain', 'enable_docker'] + ignore)

    if 'fabric' in controller:
        check_controller_fabric(personality, controller['fabric'])

    if 'blockchain' in controller:
        check_blockchain(personality, controller['blockchain'])

    if 'enable_docker' in controller:
        enable_docker = controller['enable_docker']
        if type(enable_docker) != bool:
            raise checkconfig.InvalidConfigException('invalid type "{}" for "enable_docker" in controller'.format(
                type(enable_docker)))

    return res


def check_controller_options(personality, options, ignore=[]):
    return checkconfig.check_controller_options(personality, options, ignore)


def check_hostmonitor_options(personality, options):
    checkconfig.check_native_worker_options(personality, options, ignore=['interval', 'monitors'])

    # polling interval of sensors in ms
    interval = options.get('interval', 500)
    if type(interval) not in six.integer_types:
        raise checkconfig.InvalidConfigException(
            'invalid type "{}" for "interval" in host monitor configuration (must be an integer for ms)'.format(
                type(interval)))

    monitors = options.get('monitors', {})
    if not isinstance(monitors, Mapping):
        raise checkconfig.InvalidConfigException(
            'invalid type "{}" for "monitors" in host monitor configuration (must be a dict)'.format(type(monitors)))
    for monitor in monitors:
        # FIXME: check if we know the monitor, and monitor
        # specific configuration is valid
        pass


# check native worker options of market maker workers
def check_markets_worker_options(personality, options):
    checkconfig.check_native_worker_options(personality, options, ignore=[])


# check market maker configuration items (as part of market maker workers)
def check_market_maker(personality, maker):
    maker = dict(maker)
    checkconfig.check_dict_args(
        {
            'id': (True, [six.text_type]),
            'key': (True, [six.text_type]),
            'database': (True, [Mapping]),
            'connection': (True, [Mapping]),
            'blockchain': (False, [Mapping]),
        }, maker, "market maker configuration {}".format(pformat(maker)))

    check_database(personality, dict(maker['database']))

    checkconfig.check_dict_args({
        'realm': (True, [six.text_type]),
        'transport': (True, [Mapping]),
    }, dict(maker['connection']), "market maker connection configuration")
    checkconfig.check_connecting_transport(personality, dict(maker['connection']['transport']))

    if 'blockchain' in maker:
        check_blockchain(personality, maker['blockchain'])


# check native worker configuration of maker maker workers
def check_markets_worker(personality, config):

    for k in config:
        if k not in ['id', 'type', 'options', 'makers']:
            raise checkconfig.InvalidConfigException(
                'encountered unknown attribute "{}" in XBR markets worker configuration'.format(k))

    if 'id' in config:
        checkconfig.check_id(config['id'])

    if 'options' not in config:
        raise checkconfig.InvalidConfigException('missing attribute "database" in XBR markets worker configuration')

    check_markets_worker_options(personality, config['options'])

    if 'extra' not in config['options']:
        raise checkconfig.InvalidConfigException(
            'missing attribute "options.extra" in XBR markets worker configuration')

    extra = config['options']['extra']

    if 'database' not in extra:
        raise checkconfig.InvalidConfigException(
            'missing attribute "options.extra.database" in XBR markets worker configuration')

    check_database(personality, extra['database'])

    if 'blockchain' not in extra:
        raise checkconfig.InvalidConfigException(
            'missing attribute "options.extra.blockchain" in XBR markets worker configuration')

    check_blockchain(personality, extra['blockchain'])

    makers = config.get('makers', [])

    if not isinstance(makers, Sequence):
        raise checkconfig.InvalidConfigException("'makers' items must be lists ({} encountered)\n\n{}".format(
            type(makers), pformat(config)))

    for maker in makers:
        check_market_maker(personality, maker)


_native_workers = default_native_workers()

# Override existing worker type: router workers
_native_workers.update({
    'router': {
        'class': RouterWorkerProcess,
        'worker_class': ExtRouterController,

        # check a whole router worker configuration item (including realms, transports, ..)
        'checkconfig_item': checkconfig.check_router,

        # only check router worker options
        'checkconfig_options': checkconfig.check_router_options,
        'logname': 'Router',
        'topics': {
            'starting': 'crossbar.on_router_starting',
            'started': 'crossbar.on_router_started',
        }
    }
})

# New worker type: host monitor
_native_workers.update({
    'hostmonitor': {
        'process_class': HostMonitor,
        'class': HostMonitorProcess,
        'worker_class': HostMonitor,

        # FIXME: check a whole hostmonitor configuration item
        'checkconfig_item': do_nothing,
        # FIXME: only check hostmonitor worker options
        'checkconfig_options': check_hostmonitor_options,
        'logname': 'Hostmonitor',
        'topics': {
            'starting': 'crossbar.on_hostmonitor_starting',
            'started': 'crossbar.on_hostmonitor_started',
        }
    }
})

# New worker type: XBR Market Maker ("xbrmm")
_native_workers.update({
    'xbrmm': {
        'process_class': MarketplaceController,
        'class': MarketplaceControllerProcess,
        'worker_class': MarketplaceController,
        'checkconfig_item': check_markets_worker,
        'checkconfig_options': check_markets_worker_options,
        'logname': 'XBRMM',
        'topics': {
            'starting': 'crossbar.on_xbrmm_starting',
            'started': 'crossbar.on_xbrmm_started',
        }
    }
})


class Personality(CrossbarPersonality):

    log = txaio.make_logger()

    NAME = 'edge'

    TEMPLATE_DIRS = [('crossbar', 'edge/webservice/templates')] + CrossbarPersonality.TEMPLATE_DIRS

    WEB_SERVICE_CHECKERS: Dict[str, object] = {
        'pairme': RouterWebServicePairMe.check,
        **CrossbarPersonality.WEB_SERVICE_CHECKERS
    }

    WEB_SERVICE_FACTORIES: Dict[str, object] = {
        'pairme': RouterWebServicePairMe,
        **CrossbarPersonality.WEB_SERVICE_FACTORIES
    }

    REALM_STORES: Dict[str, object] = {'cfxdb': RealmStoreDatabase, **CrossbarPersonality.REALM_STORES}

    check_controller = check_controller
    check_controller_options = check_controller_options

    check_market_maker = check_market_maker

    Node = FabricNode
    NodeOptions = NodeOptions

    native_workers = _native_workers
